#include "osihfe.h"
#include "hfe.h"
#if defined (_WIN32)
#include <sys/types.h>
#include <sys/stat.h>
#endif
#include <stdio.h>
#include <stdlib.h>
#include <memory.h>

#define BCD2BIN( a ) ( ((( a >> 4) & 0x0F) *10) + ( a & 0x0F) )
#define BIN2BCD( a ) ( ((a / 10) << 4) + a%10 )
#ifndef __min
#define __min(X,Y) ((X) < (Y) ? (X) : (Y))
#endif

#define IsValidBCD(a) (((( a >> 4) & 0x0F) <= 9) && ((a & 0x0F) <= 9) )

extern FILE* logfile;

const char szHeaderSig[] = "HXCPICFE";


/* OSI OS65D disk image info */
//
// 5.25" disk, 40 tracks, 8  sector/track,  256 bytes/sector, 11 bits/byte (8E1) =80K/disk
// OSI uses 8E1 to give a max unformatted capacity of 2272 bytes per track (see below)
// However other bit encodings  (8N1) could give up to 2500 bytes/track 
//
// The standard speed for mini-floppy drives (5") is 300 rpm. Thus one 
// rotation of the disk is 200 mS.
//
// Stated baud-rate is 125k or 125000 bits/sec and one serial byte is 11 bits 
// (1 start, 8 data, 1 parity, 1 stop).
//
// So the theoretical absolute maximum storage per track is -
//    (125000 x 0.2) / 11  =  2272 bytes  or 8.8 pages
//
// OS-65D loses a bit more because it doesn't write until 10 mS after the 
// index pulse, so -
//    (125000 x 0.19) / 11  =  2159 bytes  or 8.4 pages
// and this doesn't even allow for the length of the index pulse (a few mS)? 
// and the speed variation between drives.
//
// Data clocked out at 125000 bps, 11 bit word =11363 cps. 1Mhz clk = 88 cycles
// per character or 176 at 2Mhz	 (Floppy flux rate is 250000 fps)

//---------------------------------------------------------------------------

// 8" disk, 77 tracks, 12 sectors/track, 256 bytes/sector, 11 bits/byte (8E1/8N1) = 231K/disk
// OSI uses 8E1 to give a max unformatted capacity of 3772 bytes/track (see below)
// However other bit encodings (8N1) could give up to 3900 bytes/track
//
// The standard speed for 8" floppy dives is 360 rpm. Thus one rotation of 
// the disk is 166.6ms
//
// Stated baud rate is 250K or 250000 bits/sec and one serial byte is 11 bits
// (1 start, 8 data, 1 parity, 1 stop).
//
// So the theoretical absolute maximum storage per track is -
//    (250000 x 0.166 ) / 11 = 	3772 or 14.7 pages
//
// OS-65D loses a bit more because it doesn't write until 10 mS after the 
// index pulse, so -
//    (250000 x 0.156) / 11  =  3545 bytes  or 13.8 pages
// and this doesn't even allow for the length of the index pulse (a few mS)? 
// and the speed variation between drives.
// 
// Data clocked out at 250000 bps, 11 bit word =22727 cps. 
// floppy flux rate is 500000 fps
//---------------------------------------------------------------------------
// Track 0 format
// 10 millisecond delay after index hole
// 0,1 load address of the track in hi,lo form
// 2, page count of how much data is written on track 0, usually 8
// 3+ data
// 
// (SEE UPDATE BELOW for real timing values) 
//
// Track 1-n Format	 
// 10 ms delay after index hole
// 0,1 -a 2 byte start code $43, $57
// 2,  -BCD track number
// 3,  -track type code (always $58)
// (sector data follows)
//
// There can be any mixture of various length sectors. The total page count can not
// exceed 8 pages(12 on 8") (8*256) if more than one sector is on a track
// Each sector is written in the following format:
// -previous sector length (4 if none before) times 800 microseconds of delay
// -sector start code $76
// -sector number in binary
// -sector length (#pages) in binary
// -sector data
// -(usually has end of sector mark? $47, $53 )
//
// Directory Format
// 2 sectors (1 & 2) on track 12 (8 for 8") hold the directory information.
// Each entry requires 8 bytes. There are a total of 64 entries. The entries are
// formatted as follows:
// 0-5 ASCII 6 character filename
// 6 BCD first track of file
// 7 BCD Last track of file
//
//------------------------------------------------------------------------------

// OS65U (for Business systems)
// OS65U sector size is 3584, one sector per track on 8" floppy 
//
// Track 0 format
// 10 millisecond delay after index hole 8E1
// 0,1 load address of the track in hi,lo form
// 2, page count of how much data is written on track 0, usually 13
// 3+ data
//
// Track 1-76 format
// wait for index hole, set 8E1
// $80, tkLo, tkHi, Reset ACIA 8N1 +few ms delay
// <sometimes noise with hi bit set> $0F or $1F marker?, $E00 (3584) bytes data + 2bytes checksum (data & header)
//
// File entries 1st 16 bytes contain a copy of the diretory entry
//

// -------------------------------------------------------
// HEXDOS operating system for OSI C1 series
// Track 0 follows normal OSI boot track rules ecxept there are 9 pages of data
// loading at $0300
//
// Track 01-39
// $53, trk# (in binary) <2048 bytes of data>
//directory track is track 1
//directory consists of basic compatible listing loaded at $0B00
//@$0B00:                                    
//00,   0C ,  0B  ,  02  ,  00  ,HEXDOS, 00, 12, 0B, 09, 00 , A5, 00, 00, 00, 00, 00, 00
//00, ptrLo, ptrHi, trkLo, trkHi, <name>,00, 
// ptrLo, ptrHi points to next entry unless 0 which indicates end of list
// the last entry is a *(A5?) indicating the 1st available track
 
// DOS/65
// Track 0-n Format
// 10 ms delay after index hole
// 0,1 -a 2 byte start code $43, $57
// 2,  -BCD track number
// 3,  -track type code (always $58)
// -sector start code $76
// -sector number in binary	 (always 1)
// -sector length (#pages) in binary (always 8 (or 13 on 8"))
// -sector data	(up to 8*256, 13*256)


/*
  (from T.R. Berger in Compute's Gazette #20 Jan 1982 p136)
*** Actual OS65D disk timing ***

 The data on track 1+ commences 1ms (1000us) past index hole, about 23 bytes in time at 44us/byte
 2 bytes are written to indicate start of track. $43, $57. When the track is read, reading
 does not commence until $43 $57 have been found. Next the track number in BCD is written,
 then a stop byte $58 is written (it is never checked on read). This constitutes the track
 header. On initialization, the track is erased, then the track header is written. It is not
 rerecorded on any other read or write.
 There is a lull after track header of 6.6ms (about 149 bytes) This time differs greatly from
 the time given in the OS65D Guide!
 During a sector seek operation, a "previous sector length" number is saved "p". This value
 is set to 4 for a sector 1 seek. Then a subroutine waits p * 800us. The OS65D Guide says
 that between sector N and Sector N+1 there is a p * 800ms gap. This is not quite correct.
 After the end of sector OS65D waits p * 800us, the write function is switched on, a further
 185us is allowed to pass then the erase function is switched on. we now wait an additional
 p * 800us before starting to write data. In other words the time from the last byte of sector
 N to the first byte of sector N+1 is about p * 1600us + 185us. For sector 1 p is taken to be 4
 In all other cases "p" is the length in pages (256bytes) of the previous sector.
 This applies to 1MHz systems. On boot OS65D measures the timing on the serial port to measure
 clock speed, then a timing constant in the 1ms subroutine is set. However the timing calculation
 does not affect the 100us routine used in sector spacing. So the 100us routine is really a
 100us/T where T=the MHz of the machine. This calculation only accounts for the wait loops in
 intersector timing. There is quite a lot of linear code. A crude estimate would be to add an
 additional 30us after each sector. In other words the sector spacing is (p * 1600+215)/T us.
 where p is the number of pages in the previous sector and T is the system speed in MHz.
 All of this says there is some kind of empty space between the track header and the 1st sector.
 Each sector is completely rewritten each time it is addressed in a write operation.
 A sector is written as follows: A sector start code $76 is written, next the sector number and
 the sector length in pages "p".  The sector is verified with a read operation against the value
 in memory. Next is the data, p*256 bytes long. followed by a sector gap already covered.
 When writing sector N, the computer must read the preceeding sectors. While searchig for a sector
 it also watches for index hole again to keep from hanging on a sector read. After the start of the
 preceeding sector, the computer no longer searches for an index hole as there is not enough time
 to watch for end and prepare for the next sector.  If the index hole passes during a write operation
 the header may be overwritten rendering the track unreadable.
 So on a 1MHz machine with 8" floppies spinning at 360 RPM, one revolution takes 166,667us
 the disk data is clocked at 250KHZ each bit takes 4us. Since OSI bytes uses 11bits, 44us are 
 required per byte. If we could pack a track we could fit 3787 bytes. But it is not packed
 accounting for formatting time we use 1000ms from the index hole, The header is 4bytes long
 using 176us more. The gap to sector 1 is 4*1600+215 or 6615us, so 7791us are used between
 the index and start of sector 1. Each sector contains an integral number of pages, thus all
 sectors contain as an aggregate p pages. Each byte take 44us and there are 256 bytes per page
 thus all these pages take p*11264us. Each sector takes 5 extra bytes thus for n sectors we have 
 n * 220us. Next we must account for all the wait time after each of the n sectors.
 The wait time for one sector to the next is q * 1600+215us where q is the number of pages in
 the preceeding sector. Since we assume there are n sectors on a track there are only n-1
 spaces between sectors. If the last sector has r pages, then the preceding n-1 contain
 p-r pages altogether. Thus the total intersector wait time is 1600*(p-4)+215*(n-1)us. Finally
 we must account for the time after the last sector is written before the erase functions are 
 switched off. Write and erase continue for r*600us after the last byte is written. The write is
 switched off and erase continues for 600us more before it is switched off. This total trailing
 time is 525+600*r us.
 For 8" floppies with a 1MHz clock the total recording time for n sectors is:
 Tus=8101+12864*p -1000*r +435*n us, where p is the total numer of pages, and r is the number
 of pages in the last sector
 For 5.25" floppys clock data at 125KHz and spin at 300RPM

 8 INCH FLOPPY TIMING
 # Sectors   Total    Pages         Time      DT
             Pages 	 Last Sector
	  1		  13	    13		   162768	 3900
	  2		  13		10		   166203	  464
	  3		  13		10		   166638	   29
	  4		  12		1		   163209	 3458
	  5		  12		1		   163144	 3028
	  6		  12		1		   164079	 2588
	  7		  12		1		   164514	 2153
	  8		  12		1		   164949	 1718
	  9		  12		1		   165384	 1283
	 10		  12		1		   165819	  848
	 11		  12		1		   166254	  413
	 12		  12		1		   166689	  -22
	 t(us.) = 8101 + 12864*p - 1000*r + 435*n
	 p = number of pages, r = number of pages in last sector, n=number of sectors
	 166667us = time on one track
	 DT = time left on track

 5.25 INCH MINI FLOPPY TIMING
 # Sectors   Total    Pages         Time      DT
             Pages 	 Last Sector
								   
	1		  8			8		   193986	  6014
	2		  8			3		   199641	   359
	3		  8			4		   199296	   704
	4		  8			4		   199951		49
	5		  7			1		   179478	 20522
	6		  7			1		   180133	 19867
	7		  7			1		   180788	 19212
	8		  8			1		   205571	 -5571
	t(us.) = 8307 + 24128*p -1000*r + 435*n
	200000 us = time on one track
*/

//supported software ACIA encoding types
enum eSERFORMAT {
	s8E1 = 0,
	s8N1 = 1,
};

//index to bitmask 
const unsigned char mask[8] = { 0x01, 0x02, 0x04, 0x08, 0x10, 0x20, 0x40, 0x80 };


// FM decoding state for converting from HFE
struct FMstate
{
	eSERFORMAT eSformat; //serial word format
	int sWordIndex;		 //serial word bit index, 0 = lsb
	bool bSerWord[11];	//0...10 for 8E1
	bool bClock;		//flag we have FM clock pulse 
	bool bPastIndex;
	unsigned char data;	//current input character
	unsigned int nPos;	//bit index
	int nSampleCount;	//counting the source bits required for a data bit
	int nZeroRun;
	int nulRun;
	

};

//FM encoding state
struct FMOutData
{
	//eSERFORMAT eSformat; //serial word format
	unsigned char data;	//current output character
	unsigned int nPos;	//bit index

};

//OSI OS65D sectors can be any combination of up to 8 or 13 sectors (5.25/8") totalling 8/13 256 byte pages
// So you can have 8 1-page sectors. 2 4-page sectors , or a 1,2,1,1,3 page sector track on 5.25"
struct OSISector
{
	int nBlockLen[20]; //length of sector
	const unsigned char * pData[20];  //pointer to sector data
};

//find first occurrence of pFind in pSearch of specified lengths 
unsigned char * bStrStr( unsigned char * pSearch, unsigned char *pFind, int nSearchLen, int FindLen)
{
	if (FindLen > nSearchLen)
		return NULL;
	int pos, matchpos;
	for (pos = 0; pos + FindLen < nSearchLen; pos++)
	{
		matchpos = 0;
		if ( (pSearch[pos] ) == pFind[matchpos])
		{
			while (matchpos < FindLen)
			{
				if ( (pSearch[pos+matchpos]) != pFind[matchpos])
				{
					break;
				}
				matchpos++;
			}
			if (matchpos == FindLen)
			{
				return pSearch + pos;	
			}
		}
	}
	return NULL; 
}
// int handleData(FMstate& fmstate, int pos, bool bVerbose)
//
//return decoded character if >= 0
//return contine otherwise
//
//pos indicates position in input data for data logging
//fmstate is the current serial data state
//bVerbose enables extended data logging 
//
int handleData(FMstate& fmstate, int pos, bool bVerbose)
{
	int nRet = -1;
	int nParity = 0;
	int i;
	
	//start bit is 0, stop bit is 1, but bits are INVERTED due to ACIA

	switch(fmstate.eSformat)
	{
		default:
		case s8E1: //start bit = bit 0, data 1..8, parity 9, stop 10
				if (fmstate.sWordIndex >= 11)
				{
					//check for correct framing (inverted)
					if (fmstate.bSerWord[0] == 1 && fmstate.bSerWord[10] == 0)
					{
						nRet = 0;

						//decode serial byte
						for (int i = 8; i >= 1; i--)
						{
							
							nRet = (nRet << 1);
							if (!fmstate.bSerWord[i])
							{
								nRet |= 1;
								nParity++;
							}
						}
						// even parity - sum of data bits + parity is even
						if (!fmstate.bSerWord[9])
							nParity++;

						if (nParity & 1)
						{
							//Parity Error
							if (bVerbose)
								LogText( _T("PE @ %05lX\r\n"),  pos);
								
						}

						fmstate.sWordIndex = 0;
					}
					else
					{
						//keep shifting lsb out until framing works
						for ( i = 1; i <= 10; i++)
						{
							fmstate.bSerWord[i-1] = fmstate.bSerWord[i];
						}
						fmstate.bSerWord[10] = 0; 	
						fmstate.sWordIndex = 10;


					}
				}
			break;

		case s8N1:	//start bit = bit 0, data 1-8, stop bit 9
				if (fmstate.sWordIndex >= 10)
				{
					if (fmstate.bSerWord[0] == 1 && fmstate.bSerWord[9] == 0)
					{
						nRet = 0;
						//decode serial byte
						for (int i = 8; i >= 1; i--)
						{
							nRet = (nRet << 1);
							if (!fmstate.bSerWord[i])
								nRet |= 1;
						}
						fmstate.sWordIndex = 0;

					}
					else
					{
						//shift lsb out until framing works
						for ( i = 1; i <= 9; i++)
						{
							fmstate.bSerWord[i-1] = fmstate.bSerWord[i];
						}
						fmstate.bSerWord[9] = 0; 	
						fmstate.sWordIndex = 9;
					}	
				}



			break;
	}

	return nRet;	
	
}

void FMdecodeInitialize(FMstate& fmstate)
{
	memset( &fmstate, 0, sizeof(FMstate));
	fmstate.eSformat = s8E1;
	
}


// - decoded data may be spread across multiple calls to FMdecodestream, i.e. there may be incomplete decodes
//   when inBuf is exhausted.
// - not all input data may be consumed in one call to FMdecodestream if outBuf is too small
//
// - looks like stream of 001000100010001000100010... is clock bit with no (0) data and
//              stream of 101010101010101010101010... is clock bit with a (1) data, 2 bits HFE data to 1bit data
//                   
int FMdecodestream(const unsigned char *inBuf, int nInLen, unsigned char* outBuf, int nOutLen, FMstate& fmstate, int* pnRead, int *pnWrote, int pos, bool bVerbose)
{
//assume 1st "1" bit is clock, 2nd bit is data. Assume lsb of inBuf	is 1st bit written to FDC
//HFE FM data is 2bits for each clock or data bit
//assume OSI data is stored stored bit 0 first (lsb) in serial word format (10 or 11 bits)
//the start bit of the serial word is on disk is a '1' because serial data is INVERTED
//data bits can be spread between blocks, a clock could be at the end of one, data at the next
//but generally all data is nibble aligned

	int nRead = 0;
	int nWrote = 0;
	bool bcont = true;

	//read bits from inBuf characters, until a '1' is encountered
	//that is clock bit
	//read next bit into recv buf, until 1 serial word is recovered
	//check for serial word framing, parity etc. errors
	//write decoded serial word to output buffer
	while( nRead < nInLen && bcont)
	{
		int nChar = 0;

		if (fmstate.nPos == 0)
		{
			fmstate.data = inBuf[nRead++]; //get next byte to process
			pos++;
		}
		//while processing input byte
		while (fmstate.nPos < 8)
		{
			bool b = (fmstate.data & mask[fmstate.nPos++]) > 0;

			if (fmstate.nSampleCount++ < 0)	 //we only look at every other bit from HFE data after syncing
			{			
				continue;
			}
			
			if (fmstate.bClock)
			{
				//FM clock has been detected, now read data
				//if next 3 bits are 0, then data bit is 0
				//else if at least one of next 3 bits is 1, data is 1
				if (!b)
				{
					if (fmstate.nSampleCount < 1) //just skipping 0's
					{
						continue;
					}
				}
				//reading data
				fmstate.bSerWord[fmstate.sWordIndex++] = b;
				fmstate.bClock = false;		//got data, now wait for FM clock
				fmstate.nSampleCount = -1;	//eat next bit
				if (!b)
					fmstate.nZeroRun++;
				else 
				{
					if (fmstate.nZeroRun > 20)
					{
						if (bVerbose)
							LogText( _T(" Zeros %02d \r\n"),fmstate.nZeroRun);
						fmstate.bPastIndex = true;
					}
					fmstate.nZeroRun = 0;
				}
					

				nChar = handleData(fmstate, pos, bVerbose);  //process bit from data stream
				if (nChar >= 0 )
				{
					//are we done with a byte out?
					

					if (fmstate.bPastIndex)
					{
						
						if (bVerbose)
						{
							LogData((unsigned char) nChar);

						}						
						*outBuf++ = (unsigned char) nChar;

						if (++nWrote >= nOutLen)
						{
							//outbuf full
							bcont = false;
							break;
						}
					}
					else
					{
						if (bVerbose)
							 LogText(_T("(%02X)"), (unsigned char)  nChar);	
					}
				}
			}
			else
			{
				//we're expecting clock
				if (b)
				{					
					//have clock
					fmstate.bClock = true;
					//when expecting clock, on 1st transition, eat next bit, whatever it is
					fmstate.nSampleCount = -1;
					if (fmstate.nulRun > 100)
					{
						if (bVerbose)
							LogText( _T(" Null %02d \r\n"),fmstate.nulRun);
						
					}
					fmstate.nulRun = 0;
				}
				else
				{
					fmstate.nulRun++;
				}
			}			
		}
		fmstate.nPos = 0;

	} //while there is input

	if (pnRead)
	{
		*pnRead = nRead;
	}
	if (pnWrote)
	{
		*pnWrote = nWrote; 
	}
	return 1;

}

// Write a bitlen number of FM encoded 0's, return # character written;
// one clock pulse before each 0, 4 bits out for each 0 bit
int FMWriteZeros( int bitlen, FMOutData fmd, unsigned char* lpOut, int nOutLen)
{
    //one FMclock + 0 is 4 bits out	 0100
	unsigned char c = 0;
	int nWrote = 0;
	
	if (fmd.nPos > 0)
	{								     //00100010
		//finish writing last bit
		c = (unsigned char) ((fmd.data & 0x0F) | 0x20);	 //0010xxxx
		fmd.nPos = 0;
		fmd.data = 0;
		*lpOut++ = c;
		bitlen--;
		nWrote++;
	}

	while (bitlen > 2 && nWrote < nOutLen)
	{
		//write a bunch of pairs
		*lpOut++ = 0x22;
		bitlen -= 2;
		nWrote++;
	}
	if (bitlen > 0)
	{
		//last nibble is stored in the buffer
		fmd.data = 0x02;
		fmd.nPos = 4;

	}

	return nWrote;

}

//write 1 data bit,  FM clock & data = 4 flux bits (lsb 1st)
int WriteNibble(bool bWriteBit, FMOutData& fmd, unsigned char* lpOut)
{
	//one FM clock per data bit
	unsigned char c = 0;
	int nWrote = 0;

	if (bWriteBit)
		c = (unsigned char)( (fmd.nPos > 0) ? 0xA0 : 0x0A); //1010 
	else				//					  ~~ clock
		c = (unsigned char)((fmd.nPos > 0) ? 0x20 : 0x02); //0010
	if (fmd.nPos > 0)					 //	~~	 data
	{	
		//finish writing last bit
		*lpOut++ = (unsigned char) (fmd.data | c);
		fmd.nPos = 0;
		fmd.data = 0;

		nWrote++;
	}
	else
	{
		fmd.nPos = 4;
		fmd.data = c;
	}
	return nWrote;
}

// Write a byte in FM + 8E1 format (11 bits data, 44 output bits, 5.5 bytes) 
int FMWrite8E1(unsigned char data, FMOutData& fmd, unsigned char* lpOut, int OutLen)
{
	int nWrote = 0;
	int nParity = 0;
	(void)OutLen;   //Do nothing

	//(data is inverted upon write)
	//write start bit
	nWrote = WriteNibble(1, fmd, lpOut);

	//write data bits
	for (int i = 0; i < 8; i++)
	{
		bool bSet = (data & mask[i]) > 0;

		nWrote += WriteNibble( !bSet, fmd, lpOut+nWrote);
		if (bSet)
			nParity++;
	}
	//write parity
	nWrote += WriteNibble( !(nParity & 1), fmd, lpOut+nWrote);

    //stop bit
	nWrote += WriteNibble( 0, fmd, lpOut+nWrote);
	
	return nWrote;

}

// Write a byte in FM + 8N1 format (10 bits)
int FMWrite8N1(unsigned char data, FMOutData& fmd, unsigned char* lpOut, int OutLen)
{

	int nWrote = 0;
	(void)OutLen;   //Do nothing


	//(data is inverted upon write)
	//write start bit
	nWrote = WriteNibble(1, fmd, lpOut);

	//write data bits
	for (int i = 0; i < 8; i++)
	{
		bool bSet = (data & mask[i]) > 0;

		nWrote += WriteNibble( !bSet, fmd, lpOut+nWrote);
	}

    //write 1 stop bit
	nWrote += WriteNibble( 0, fmd, lpOut+nWrote);
	
	return nWrote;

}

//locate start of track data for OS65D, Hexdos, os65U
//there may be garbage bytes before expected  data
int OSISeekTrackStart( const unsigned char* pTrack, int nLen, DISKTYPE dt /*= DISK_UNKNOWN*/)
{
	int i = 0;
	int guess = -1;
	
	if (!pTrack)
		return -1;

	if (nLen > 32)  //limit search to reasonable length
		nLen = 32;

	while (i+4 <= nLen)
	{
		if (pTrack[i] == 0x43 && pTrack[i+1] == 0x57 && pTrack[i+3] == 0x58 && dt != OS65U8) //OS65D
		{
			unsigned char c = pTrack[i+2];
			int nTrack = BCD2BIN(c); //BCD to decimal
			if (nTrack > 79)
			{
				//not valid OS65D track# (5.25 max = 39, 8" = 77)
				guess = i;
			}
			else
			{
				return i;
			}	
		}
		else if (pTrack[i] == 0x57 && pTrack[i+1] < 40 && dt != OS65U8)
		{	//this detection is too ambiguous, one wayward byte and 
			//hexdos
			return i;
		}
		else if (pTrack[i] == 0x80 && pTrack[i+1] == 0)
		{
			int nTrack = pTrack[i+2];
			if (nTrack > 80)
			{
				//not valid track for OS65U
			}
			else
			{
				guess = i;
				if (dt == OS65U8) //if we're sure we are looking for this, end!
					break; 
			}
		}
		
		i++;
	}

	return guess;
}

//return array of OSI OS65D/OS65U/HexDos/Dos65 sector lengths & pointers to sector data
int OSICountSectors(const unsigned char * pTrack, int nTrkLen, OSISector& sect)
{
	memset(&sect, 0, sizeof(OSISector));
	int nCount = 0;

	if (!pTrack)
		return 0;

	const unsigned char *p = pTrack;
	int i = 0;

	i = OSISeekTrackStart(p, nTrkLen);

	if (i >= 0 && pTrack[i] == 0x43)
	{
		i=i+4; //skip start of track

		// $76, sect#, sect-len 
		while (i+5 < nTrkLen)
		{
			if (pTrack[i] == 0x76) //start of sector marker
			{
				int nSect = pTrack[i+1];
				int nSectLen = pTrack[i+2];
				if (nSect < 13 && nSect > 0 && nSectLen <= 13 && nSectLen > 0 )
				{
					//looks legit
					sect.nBlockLen[nCount] = nSectLen;
					sect.pData[nCount] = (pTrack + i + 3);

					i = i + nSectLen*256 + 3;
					nCount++;
					//now expect $47,$53 

					if (i+1 > nTrkLen)
					{
						break;
					}
					if ( pTrack[i] == 0x47 && pTrack[i+1] == 0x53)
					{
						//end of sector mark sometimes present
						i++;

						//now there could be more sectors or 0's or whatever
					}

				}
				
			}
			i++;
		}
		
		
	}
	else if (pTrack[i] == 0x80 && pTrack[i+1] == 0x00)
	{
		//OS65U
		sect.nBlockLen[0] = 0xE03;
	    
		i+=3;
		if (pTrack[i] & 0x80) //skip possible noise
			i++;
		sect.pData[0] = (pTrack + i);
		nCount = 1;
	}
	else if (pTrack[i] == 0x57 && pTrack[i+1] < 40)
	{
		//hexdos
		sect.nBlockLen[0] = 2048;
		i+=2;
		sect.pData[0] = (pTrack + i);
		nCount = 1;
	}
	
	return nCount; 
}

// pTrack points to OSI track data of TRKSIZ8 or TRKSIZ5 bytes, or NULL to create blank track
// DISKTYPE specifies disk type & thus size of pTrack
// nTrack specifies OSI track #, track 0 is handled differently

int FMEncodeOSITrack( unsigned char * pTrack, DISKTYPE dt, int nTrack,  unsigned char * lpOut, int nOutLen, bool bWrite8As5)
{
	//Convert from WinOSI disk image to HFE disk image
	//must replace timing delays lost in OSI dump process
	//
	// timing after index hole, timing between encoding changes (8E1/8N1 OD65U)
	// timing before sectors OS65D
	// must extract clean data from WinOSI image

#define DISK8INDEX 800		  //in flux nibbles (clk+0)
#define DISK5INDEX 1478		  //in flux nibbles
#define DISK5TARGETLEN 17000  //in storage bytes, not flux nibbles	based on kyroflux sample image
#define DISK8TARGETLEN 20898  //in storage bytes, not flux nibbles	based on kyroflux sample image

	if (!lpOut || nOutLen < 1)
		return -1;

	int nInLen = 0;
	int nPos = 0;
	int fpm = 250;		//flux per millisecond 
	int nDisk8Index = DISK8INDEX;
	FMOutData fmd;

	//fmd.eSformat = s8E1;
	fmd.data = 0;
	fmd.nPos = 0;

	//track buffer needs up to 
	// 500000 bps * 166.6ms	= 83333 flux per track.	(data clock rate 250K)
	//  @ 2 flux per bitcell (8bits of memory) = 41666 bytes for 8" track
	// 250000 bps * 200ms = 50000 flux per track.
	// @ 2 flux per bitcell = 25000 bytes for 5.25" track
	// or so we think

	//index pulse width on 8" drive is ~1.7ms so about 850 zeros
	//from OSI formatted disk on HxC 5.25" has ~1478 zeros  (expected less??)
	//from OSI formatted disk on HxC 8" has ~757 zeros
	//fron Kryoflux image, 8" drive has about 250 zeros	(TOO LOW)
	OSISector sect;
	memset(&sect, 0, sizeof(OSISector));
	
	switch (dt)
	{
		case OS65D5:
		case HEXDOS:
		case DOS655:
		case EMPTY5:
		case EMPTY5Z:
			nInLen = TRKSIZ5;
			break;

		case OS65D8:
		case DOS658:
		case OS65U8:
		case EMPTY8:
		case EMPTY8Z:
			nInLen = TRKSIZ8;
			fpm =500;
			break;
		default:
			break;
		
		
	}
	if (bWrite8As5)
	{
		nDisk8Index = DISK5INDEX;
	}

	if (dt == EMPTY5 || dt == EMPTY8 || dt == EMPTY5Z || dt == EMPTY8Z)
	{
		nPos =  FMWriteZeros( ( (dt == EMPTY5 || dt == EMPTY5Z) ? DISK5TARGETLEN : DISK8TARGETLEN)*2 , fmd, lpOut, nOutLen);
		if (dt == EMPTY5Z || dt == EMPTY8Z)
			memset(lpOut, 0, nPos);
		
		return nPos;
	} 

	if (nTrack == 0)
	{
		//special handling for track 0
		if (dt == OS65D5 || dt == HEXDOS)
		{
			//write leader
			nPos = FMWriteZeros( DISK5INDEX, fmd, lpOut, nOutLen);
			int nLen = 8*256 + 3;
			if (dt == HEXDOS)
				nLen += 256;
			//write data
			for (int i = 0; i <	nLen && i < nInLen; i++)
			{
				if (pTrack)
					nPos += FMWrite8E1(pTrack[i], fmd, lpOut+nPos, nOutLen-nPos);
				else
					nPos += FMWrite8E1(0, fmd, lpOut+nPos, nOutLen-nPos);

			}
			//write track filler
			nPos +=  FMWriteZeros( (DISK5TARGETLEN - nPos)*2 , fmd, lpOut+nPos, nOutLen-nPos);
		}
		else //8"
		{
			//write leader
			nPos = FMWriteZeros( nDisk8Index, fmd, lpOut, nOutLen);
			int nLen = 8*256;
			
			if ( pTrack)
			{
				if (pTrack[2] > 8 && pTrack[2] < 14)
					nLen = pTrack[2] * 256;
				nLen += 3;
			}
			else
				nLen = 3 + 256*13;

			for (int i = 0; i <	nLen && i < nInLen; i++)
			{
				if (pTrack)
					nPos += FMWrite8E1(pTrack[i], fmd, lpOut+nPos, nOutLen-nPos);
				else
					nPos += FMWrite8E1(0, fmd, lpOut+nPos, nOutLen-nPos);
			}
			//write track filler
			nPos +=  FMWriteZeros( (DISK8TARGETLEN - nPos) *2, fmd, lpOut+nPos, nOutLen-nPos);


		}

	}
	else
	{
		if (dt == HEXDOS)
		{
			int nSectors = 0;
			if (pTrack)
				nSectors = OSICountSectors( pTrack, TRKSIZ5 , sect);
		
			//write leader & track id
			nPos = FMWriteZeros( DISK5INDEX , fmd, lpOut, nOutLen);
			nPos += FMWrite8E1(0x57, fmd, lpOut+nPos, nOutLen-nPos);
			nPos += FMWrite8E1((unsigned char)nTrack, fmd, lpOut+nPos, nOutLen-nPos);
			 
			if (nSectors > 0) 
			{
				const unsigned char * p = sect.pData[0];

				//write sector data
				for (int i = 0; i < sect.nBlockLen[0]; i++)
				{
					if (p)
						nPos += FMWrite8E1(*p++, fmd, lpOut+nPos, nOutLen-nPos);
					else
						nPos += FMWrite8E1(0, fmd, lpOut+nPos, nOutLen-nPos);
				
				}

				//write trailing bits or padding
				nPos += FMWriteZeros( (DISK5TARGETLEN - nPos) * 2, fmd, lpOut+nPos, nOutLen-nPos);

			 }
			 else
			 {		//no sectors, unknown or blank formatting

				const unsigned char * p = pTrack ? pTrack+2 : NULL; //skip track marker

				//write unknown track data, no special formatting
				for (int i = 0; i <  2048; i++)
				{
					if (p)
						nPos += FMWrite8E1(*p++, fmd, lpOut+nPos, nOutLen-nPos);
					else
						nPos += FMWrite8E1(0, fmd, lpOut+nPos, nOutLen-nPos);

				}
				//write trailing bits or padding
				nPos += FMWriteZeros( (( DISK5TARGETLEN ) - nPos)*2, fmd, lpOut+nPos, nOutLen-nPos);


			}

		}
		else if (dt == OS65D5 || dt == OS65D8)
		{
		    //OS65D
			int nSectors = 0;
			if (pTrack)
				nSectors = OSICountSectors( pTrack, dt == OS65D5 ? TRKSIZ5 : TRKSIZ8, sect);

			int nCount = 0;
			int lastlen = 1;
			int nBCDTrk =  BIN2BCD(nTrack); 
			
			//write leader & track id
			nPos = FMWriteZeros( dt == OS65D5 ? DISK5INDEX : nDisk8Index, fmd, lpOut, nOutLen);
			nPos += FMWrite8E1(0x43, fmd, lpOut+nPos, nOutLen-nPos);
			nPos += FMWrite8E1(0x57, fmd, lpOut+nPos, nOutLen-nPos);
			nPos += FMWrite8E1((unsigned char)nBCDTrk, fmd, lpOut+nPos, nOutLen-nPos);
			nPos += FMWrite8E1(0x58, fmd, lpOut+nPos, nOutLen-nPos);
			 
			 
			if (nSectors > 0)
			{
				while (nCount < nSectors && nOutLen-nPos > 0)
				{
					//fpm = flux per millisecond / 10 = flux per microsecond
					//now delay = previous sector length (or 4 if first) times .8ms delay
					
					//UPDATED: at 1MHz CPU clock delay after index hole is 1000us
					//delay between track header and 1st sector is 6600us !
					//delay between sectors is previous sector length in pages * 1600us + 185us
					
					int nDelay = (int)((fpm * lastlen));  //now a little longer	was (fpm * lastlen)/1.25

					nPos += FMWriteZeros( nDelay, fmd, lpOut+nPos, nOutLen-nPos);

					//write sector header 0x76, sect#, size
					nPos += FMWrite8E1(0x76, fmd, lpOut+nPos, nOutLen-nPos);
					nPos += FMWrite8E1((unsigned char)(nCount+1), fmd, lpOut+nPos, nOutLen-nPos);
					nPos += FMWrite8E1((unsigned char)sect.nBlockLen[nCount], fmd, lpOut+nPos, nOutLen-nPos);
					const unsigned char * p = sect.pData[nCount];

					//write sector data
					for (int i = 0; i < sect.nBlockLen[nCount] * 256; i++)
					{
						if (p)
							nPos += FMWrite8E1(*p++, fmd, lpOut+nPos, nOutLen-nPos);
						else
							nPos += FMWrite8E1(0, fmd, lpOut+nPos, nOutLen-nPos);
					
					}
					//write end of sector marker
					nPos += FMWrite8E1(0x47, fmd, lpOut+nPos, nOutLen-nPos);
					nPos += FMWrite8E1(0x53, fmd, lpOut+nPos, nOutLen-nPos);

					lastlen = sect.nBlockLen[nCount];
					nCount++;


				}
				//write trailing bits or padding
				nPos += FMWriteZeros( ((dt == OS65D5 ? DISK5TARGETLEN : DISK8TARGETLEN) - nPos) * 2, fmd, lpOut+nPos, nOutLen-nPos);


			 }
			 else
			 {		//no sectors, unknown or blank formatting

				int nDelay = fpm * lastlen;
				const unsigned char * p = pTrack ? pTrack+4 : NULL; //skip track marker
				nPos += FMWriteZeros( nDelay, fmd, lpOut+nPos, nOutLen-nPos);
				//write unknown track data, no special formatting
				for (int i = 0; i < (dt == OS65D5 ? SECTORS5 : SECTORS8 ) * 256; i++)
				{
					if (p)
						nPos += FMWrite8E1(*p++, fmd, lpOut+nPos, nOutLen-nPos);
					else
						nPos += FMWrite8E1(0, fmd, lpOut+nPos, nOutLen-nPos);
				}

				//write trailing bits or padding
				nPos += FMWriteZeros( ((dt == OS65D5 ? DISK5TARGETLEN : DISK8TARGETLEN) - nPos)*2, fmd, lpOut+nPos, nOutLen-nPos);


			}
		}
		else
		{	//OS65U
			//one 3584 byte sector, encoding switches mid-track
			//find start of sector data	within a reasonable about of bytes
			int i = 0;
			bool bFound = false;
			if (pTrack)
			{
				while ( i < 30 && i < nOutLen )	 //
				{
					if (pTrack[i] == 0x80) 
					{
						if (pTrack[i+1] == 0x00)
						{
							if (pTrack[i+2] == nTrack)
							{
								bFound = true;
								i +=3;
								break;
							}
						} 
					}
					i++;
				}
			}
			else
			{
				bFound = true;
			}

			if (bFound)
			{
				if ( pTrack && (pTrack[i] & 0x80) ) //skip possible noise byte, usually $FF, next data usually 0 or 1F
					i++;    
			
				 int chk = 0x80 + nTrack + 0x1F;
				//write leader
				nPos = FMWriteZeros( nDisk8Index, fmd, lpOut, nOutLen);
				//write track header
				nPos += FMWrite8E1(0x80, fmd, lpOut+nPos, nOutLen-nPos);
				nPos += FMWrite8E1(0x00, fmd, lpOut+nPos, nOutLen-nPos);
				nPos += FMWrite8E1((unsigned char)nTrack, fmd, lpOut+nPos, nOutLen-nPos);
				
				//delay after switching encoding -- how much?
				nPos += FMWriteZeros( 44, fmd, lpOut+nPos, nOutLen-nPos);
				
				for (int j = 0; j < 0xE03; j++)	 //write sector	+ checksum 
				{
					if (pTrack)
						nPos += FMWrite8N1(pTrack[i+j], fmd, lpOut+nPos, nOutLen-nPos);	
					else
					{
						if (nTrack > 0 && j == 0)
							nPos += FMWrite8N1((unsigned char)0x1F, fmd, lpOut+nPos, nOutLen-nPos);
						else if (nTrack > 0 && j == 0xE01)
							nPos += FMWrite8N1((unsigned char) ((chk >> 8) & 0xFF), fmd, lpOut+nPos, nOutLen-nPos);
						else if (nTrack > 0 && j == 0xE02)
							nPos += FMWrite8N1((unsigned char) (chk & 0xFF), fmd, lpOut+nPos, nOutLen-nPos);
						else
							nPos += FMWrite8N1(0, fmd, lpOut+nPos, nOutLen-nPos);	
					}
				}
				//write trailing bits if any
				nPos += FMWriteZeros( (DISK8TARGETLEN - nPos)*2, fmd, lpOut+nPos, nOutLen-nPos);

			}

		}

	}



	return nPos;
}

#define TRACKBUFSIZE 32768  //size of temp track buffer

//Lets create an HFE file from an OSI disk image.in memory

// pDisk is source OSI disk memory image, or NULL to create blank image
// dt specifies source format and desired target size/type
// sides indicates target # of sides (usually 2)
// ppData is returned HFE disk in memory (caller must free)
// pnLen returns size of HFE image in memory
//
// return true if OK
bool CreateHFEImage( unsigned char * pDisk, DISKTYPE dt, int nSides, unsigned char** ppData, int* pnLen, bool bWrite8As5 )
{
	int nTracks = 0; //40 for 5.25, 77 for 8"
	int nTrackLen = 0;
	if (nSides < 1 || nSides > 2)
		nSides = 1;

	switch (dt)
	{
		case OS65D5:
		case HEXDOS:
		case DOS655:
		case EMPTY5:
		case EMPTY5Z:
			nTracks = MAXTRK5+1;
			nTrackLen = TRKSIZ5;
			break;			 
		case OS65D8:
		case OS65U8:
		case DOS658:
		case EMPTY8:
		case EMPTY8Z:
			nTracks = MAXTRK8+1;
			nTrackLen = TRKSIZ8;
			break;
		default:
			return false;
	}

	unsigned char * pTrackBuf = (unsigned char *) malloc(TRACKBUFSIZE); //buffer for encoding track

	//How big?
	unsigned int dwHeader = 512; //sizeof(picfileformatheader); 
	unsigned int  dwTrackTable =  (sizeof(pictrack) * nTracks);
	 dwTrackTable = ((dwTrackTable + 511) / 512) * 512; //padded table size
	//how big are tracks?
	//5ms index pulse, + data
	//5.25 = 300RPM, or one rotation every 200ms
	//transfer rate is 250Kb
	// max flux per track is 50000 	(25000 bytes)
	// double sided = 100000 /8 = 12500 bytes
	//at 247,000 bps, 1 track can not exceed 49400 bits, 6175 bytes 

	//8" = 360RPM or one rotation every 166.6ms 
	// transfer rate is 500Kb
	// max flux / track is 83300
	// double sided - 166600 /8 = 20825 bytes

	int nTrackTotalSize = 0;
	

	//figure out the size of the encoded tracks
	for (int i = 0; i < nTracks; i++)
	{
		unsigned char * pTrack = NULL;
		if (pDisk)
			pTrack = pDisk + (nTrackLen * i);
		memset(pTrackBuf, 0, TRACKBUFSIZE);
		int nLen = FMEncodeOSITrack( pTrack, dt, i,  pTrackBuf, TRACKBUFSIZE, bWrite8As5);

		//track data is padded
		int nPadded = (nLen + 511) / 512 * 512;
		
		nTrackTotalSize += (nPadded * nSides); //2 = double sided ignored data
	}

	//let make an HFE image;
	unsigned char *pHFE = (unsigned char *) malloc( dwHeader + dwTrackTable + nTrackTotalSize);
	if (pHFE)
	{
		memset(pHFE, 0, dwHeader + dwTrackTable + nTrackTotalSize);
		memset(pHFE, 0xFF, dwHeader+dwTrackTable);


		picfileformatheader* pHead = (picfileformatheader*) pHFE;
		memcpy(pHead->HEADERSIGNATURE, szHeaderSig, sizeof(szHeaderSig));
		pHead->formatrevision = 0;
		pHead->number_of_track = (unsigned char)nTracks;
		pHead->number_of_side = (unsigned char)nSides;
		pHead->track_encoding = 0xFF;
		pHead->bitRate = (unsigned short)(dt == OS65D5 ? 250 : 500);
		pHead->floppyRPM = (unsigned short)(dt == OS65D5 ? 300 : 360);
		pHead->floppyinterfacemode = GENERIC_SHUGGART_DD_FLOPPYMODE;
		pHead->dnu = 1; //(?)
		pHead->track_list_offset = 1; //(=512)
		pHead->write_allowed = 0xFF;
		pHead->single_step = 0xFF;
		pHead->track0s0_altencoding = 0xFF;
		pHead->track0s0_encoding = 0xFF;
		pHead->track0s1_altencoding = 0xFF;
		pHead->track0s1_encoding = 0xFF;

		pictrack* TrkLst = (pictrack*)(pHFE + 0x200);


		unsigned char *pHFETracks = (unsigned char *) (pHFE + dwTrackTable + dwHeader);
		
		//make a list of track lengths
		for (int i = 0; i < nTracks; i++)
		{
			unsigned char * pTrack = NULL;
			if (pDisk)
				pTrack = pDisk + nTrackLen * i; 
			int nLen = FMEncodeOSITrack( pTrack, dt, i,  pTrackBuf, TRACKBUFSIZE, bWrite8As5);

			memset(pTrackBuf+nLen, 0, TRACKBUFSIZE-nLen); //erase rest of track data 

			//now interleave blocks to create double sided data
			TrkLst[i].offset = (unsigned short) ((pHFETracks - pHFE) / 512);
			TrkLst[i].track_len = (unsigned short) (nLen*nSides);

			int nPadded = (nLen*nSides + 511) / 512 * 512;

			int nBlocks = (nLen + 255) / 256;

			if (nSides == 2) //interleave blocks
			{
				unsigned char *p = pTrackBuf;
				unsigned char *pOut = pHFETracks;
				for (int j = 0; j < nBlocks; j++)
				{
					memcpy(pOut, p, 256);	//copy side 1 block
					pOut += 256;
					p += 256;
					memset(pOut, 0x11, 256); //fill side 2 with FM 0's (blank)
					pOut += 256; 	
					
				}
			}
			else   //single data block
			{
				memcpy(pHFETracks, pTrackBuf, nLen);
			}

			
			pHFETracks += nPadded;
		}

		if (pnLen )
			*pnLen = dwHeader + dwTrackTable + nTrackTotalSize;

		if (ppData)
		{
			*ppData = pHFE;
		}
		else
		{
			free(pHFE);
			pHFE = NULL;
		}
		return true;
	}

	free(pTrackBuf);
	pTrackBuf = NULL;

	return false;
}

//Determine OSI disk type from supplied disk image in memory
// Dos/65 has normal OSI track on track 0, otherwise looking at data is the only way to see
//
DISKTYPE IdentifyDiskType( unsigned char *pMem, int nReadBytes)
{
	DISKTYPE dt = DISK_UNKNOWN;
	if (!pMem || nReadBytes < 2048)
		return dt;
	
	int nTrackLen = TRKSIZ5;

	//is it OSIdisk?  (allow for slop)
	if ( (nReadBytes >  TRKSIZ5 * 41) && (nReadBytes < TRKSIZ8 * 81) )
	{
		nTrackLen = TRKSIZ8;
	}
	if (nTrackLen > nReadBytes)	//too small
		return dt;

	OSISector sect;
	int nFound = 0;
	int i;
	//OS65D 0x43, 0x57, Trk, 0x58, ..., $76, #sec , sectLen,
	//HexDos $53, trk#  

	for (i = 0; i < 20 && nFound < 4; i++) //start at track 1 for a while
	{
		DISKTYPE dSearch = DISK_UNKNOWN;
		int offset = i * nTrackLen;
		unsigned char * pTrack = pMem + i * nTrackLen;

		int nStart =  OSISeekTrackStart( pTrack, nReadBytes-offset);

		if (i == 0 && nStart >= 0)
		{
			//check track 0	for DOS/65 sector

			unsigned char match[] = {  0x43, 0x57, 0x00, 0x58 };
			if ( !memcmp(pTrack+nStart, &match[0], sizeof(match)/sizeof(match[0])) )
			{ 
				//found OS65D normal track header on trk 0, which is usually reserved as boot trk
				//except on DOS/65
				int nCount = OSICountSectors( pTrack, nTrackLen , sect);
				if (nCount == 1)
				{
					if (sect.nBlockLen[0] == 8)
					{
						dt = DOS655;
						nFound = 4;
		
					}
					else if (sect.nBlockLen[0] == 13)
					{
						dt = DOS658;
						nFound = 4;
					}	
				}
			}
		} 
		else if (nStart >= 0)
		{
			if (pTrack[nStart] == 0x43 && pTrack[nStart+1] == 0x57 && 
				pTrack[nStart+2] == (unsigned char) BIN2BCD(i))
			{
				
				int nCount = OSICountSectors( pTrack, nTrackLen , sect);

				if (nCount > 0)
				{
					if (nTrackLen == TRKSIZ8)
						dSearch = OS65D8;
					else
						dSearch = OS65D5;
				}
			}
			if (pTrack[nStart] == 0x57 && pTrack[nStart+1] == i)
			{
				dSearch = HEXDOS;
			}

			if ( pTrack[nStart] == 0x80 && pTrack[nStart+1] == 0x00 && pTrack[nStart+2] == i )
			{
				dSearch = OS65U8;
			}
			if (dSearch != DISK_UNKNOWN)
			{

				if (dt == dSearch)
				{
					nFound++;
				}
				else if (nFound < 3)
				{
					dt = dSearch;
					nFound = 1;
				}
			}
		}

	}

	return dt;
} 

//simply check decoded data for known OSI disk formatting
DISKTYPE DetermineOSIDiskTypeFromHFE(unsigned char * pHFEDisk, int nHFELen, int nSide /*=0*/)
{
	DISKTYPE dt = DISK_UNKNOWN;
	
	if (!pHFEDisk || nHFELen < 1000 || nSide > 2)
		return DISK_ERROR;

	picfileformatheader *pHFE = (picfileformatheader*) pHFEDisk;

	if (pHFE->number_of_track > 80)
	{
		LogText( _T("HFE file has too many tracks"));

		return DISK_ERROR;

	}
	if (pHFE->track_encoding != UNKNOWN_ENCODING)
	{
		LogText( _T("HFE file encoding is not UNKNOWN_ENCODING"));

		return DISK_ERROR;
		
	}
	if (pHFE->number_of_side < 1 || pHFE->number_of_side > 2)
	{
		LogText( _T("HFE file invalid number of sides"));

		return DISK_ERROR;

	}
	if (nSide > pHFE->number_of_side )
		nSide = pHFE->number_of_side; 

	FMstate fm;
	void *pEnd = pHFEDisk + nHFELen;
	unsigned char* pTrackBuf = (unsigned char *) malloc (TRKSIZ8);
	if ( !pTrackBuf)
		return DISK_ERROR;
	memset(pTrackBuf, 0, TRKSIZ8); 
	
	pictrack* trkLst = 	(pictrack*) (pHFEDisk + pHFE->track_list_offset * 0x200);
	if (((void *) &trkLst[0]) < (pEnd))
	{
		int nFound = 0;

		for (int i = 0; i < pHFE->number_of_track && nFound < 4; i++)	 //lets skip track 0
		{
		    DISKTYPE dSearch = DISK_UNKNOWN;
			unsigned char * pData = pHFEDisk + (trkLst[i].offset * 0x0200);
			int pos = 0;
			int nDataLen = trkLst[i].track_len;
			int nOutPos = 0;

			int loop = 0;


			//new track, reset decoder
			FMdecodeInitialize(fm);


			while (loop < nDataLen)
			{	
				unsigned char * inBuf = pData + loop;

				//we're reading size 2
				if (nSide == 2)
					loop += 256;

				int nInLen = 256;
				if (nInLen + loop > nDataLen)
				{
					nInLen = nDataLen - loop;
				}

				int nRead = 0;
				int nWrote = 0;

				FMdecodestream(inBuf, nInLen, pTrackBuf+nOutPos, TRKSIZ8-nOutPos,fm, &nRead, &nWrote, pos+loop, false);		
				
				if (nWrote > 0)
					nOutPos += nWrote;
				if (nWrote >= TRKSIZ8)
				{
					//handle full output buffer
					break;
				}

				if (nSide == 2)
					loop += 256; //add one page to point to side1, updated again above
				else
					//input data is stored in 0x100 chunks, alternating between side 1 & 2
					loop += (pHFE->number_of_side ? pHFE->number_of_side * 256 : 256  );

			
			}
			if (nOutPos == 0) //no data encoded on track
			{
				dSearch = nDataLen > (DISK5TARGETLEN*2 + 5000) ? EMPTY8 : EMPTY5;
			}

			int offset =  OSISeekTrackStart( pTrackBuf, nOutPos);
			if (offset >= 0)
			{
				if (pTrackBuf[offset] == 0x43 && pTrackBuf[offset+1] == 0x57)
				{
					if (  pTrackBuf[offset+2] < 0x80 && ( abs(i- BCD2BIN( pTrackBuf[offset+2] )) <= 3) )
					nFound++;

					if (pHFE->number_of_track > 40)
						dSearch = OS65D8;
					else
						dSearch = OS65D5;
					if (i == 0 && (dt == OS65D8 || dt == OS65D5)) //OSI sector on tk0, probably DOS/65
					{
						OSISector sect;
						memset(&sect, 0, sizeof(OSISector));
						
						if (dSearch == OS65D5 && OSICountSectors( pTrackBuf, nOutPos , sect) == 1)
						{
							if ( sect.nBlockLen[0] == 8 )  //DOS65 always single sector of 8 or 13 blocks
							{
								dt = DOS655;
								nFound = 4;
							}
							else if (sect.nBlockLen[0] == 13)
							{	
								dt =  DOS658;
								nFound = 4;
							}
						}
					}

					
				}
				if (pTrackBuf[offset] == 0x57 && pTrackBuf[offset+1] == i)
				{
					if (pHFE->number_of_track <= 40)
					{
						nFound++;
						dSearch = HEXDOS;
					}

				}
				if ( pTrackBuf[offset] == 0x80 && pTrackBuf[offset+1] == 0x00 && (abs(pTrackBuf[offset+2]-i) <= 3) )
				{
					//we've found OS68U disk type which require different decoding
					
					dSearch = OS65U8;
					nFound++;

					//restart decode of mixed 8E1/8N1
					unsigned char * pData = pHFEDisk + (trkLst[i].offset * 0x0200);
					int pos = 0;
					int nDataLen = trkLst[i].track_len;
					int nOutPos = 0;

					int loop = 0;


					loop = 0;
					int nRead = 0;
					int nWrote = 0;

					//reset decoder
					FMdecodeInitialize(fm);

					int nLen = 3; //decode 1st 3 bytes
					bool bHeader = true;
					while (loop < nDataLen)
					{	
						unsigned char * inBuf = pData + loop;

						//we're reading size 2
						if (nSide == 2)
							loop += 256;

						int nInLen = 256;
						if (nInLen + loop > nDataLen)
						{
							nInLen = nDataLen - loop;
						}
						if (bHeader)
							nLen = 3 - nOutPos;
						else
							nLen = TRKSIZ8-nOutPos;

						nRead = 0;
						nWrote = 0;

						FMdecodestream(inBuf, nInLen, pTrackBuf+nOutPos, nLen,fm, &nRead, &nWrote, pos+loop, false);		

						if (nWrote > 0)
							nOutPos += nWrote;
						if (nWrote >= nLen)
						{
							if (bHeader)
							{
								FMdecodeInitialize(fm);
								fm.eSformat = s8N1;
								bHeader = false;
								nLen = TRKSIZ8-nOutPos;
								//decode rest of input buffer
								FMdecodestream(inBuf+nRead, nInLen-nRead, pTrackBuf+nOutPos, nLen,fm, &nRead, &nWrote, pos+loop, false);		

							}
							else
							{
								//handle full output buffer
								break;

							}
						}

						if (nSide == 2)
							loop += 256; //add one page to point to side1, updated again above
						else
							//input data is stored in 0x100 chunks, alternating between side 1 & 2
							loop += (pHFE->number_of_side ? pHFE->number_of_side * 256 : 256  );

					
					}

				}
			}
			if (dSearch != DISK_UNKNOWN)
			{

				if (dt == dSearch)
				{
					nFound++;
				}
				else if (nFound < 3)
				{
					dt = dSearch;
					nFound = 1;
				}
			}


		}


	}	
	
	
	free(pTrackBuf);

	
	return dt;	
}

//convert HFE file to OSI Disk
#define FLOPPY_DATA_SIZE 0x60000

//Lets create an OSI dump image from an OSI HFE disk image.in memory
//
// Decode an HFE image, emulating OSI decoding hardware, determines target type from source HFE image
// pHFEDisk - memory image of HFE file
// nHFELen - and its size
// nSide  - source side (0 or 1/ or 2)
// ppOSI - returned OSI memory image - caller must free
// pnOSILen - and its size
bool CreateOSIImage( unsigned char * pHFEDisk, int nHFELen, int nSide /*=0*/, unsigned char** ppOSI, int* pnOSILen, bool bVerbose )
{

	unsigned char * OSIDisk = NULL;
	pictrack* trkLst = NULL;
	picfileformatheader *pHFE = (picfileformatheader*) pHFEDisk;;
	unsigned char * pEnd = pHFEDisk + nHFELen;
	FMstate fm;
	int nTrackSize = TRKSIZ5;
	DISKTYPE dt = DISK_UNKNOWN;
	int nTrkOffset = 0;

	if (nSide < 0 || nSide > 2)
		return false;
	if (nSide == 0)
		nSide = 1;


	if (!pHFEDisk || !ppOSI)
		return false;


	if (memcmp(pHFEDisk, szHeaderSig, 8))
	{
		//invalid data
		LogText( _T("HFE file header not found in input file"));
		goto abort;
	}


	if (pHFE->number_of_track > 80 || pHFE->track_encoding != UNKNOWN_ENCODING || pHFE->number_of_side < 1 || pHFE->number_of_side > 2)
	{
		if (pHFE->number_of_track > 80)
		{
			LogText( _T("Unsupported HFE file type"));
			goto abort;
		}
		else
		{
			pHFE->number_of_track = 80;  //ignore long tracks
		}

	}


	if (pHFE->number_of_track > 40)
		nTrackSize = TRKSIZ8;


	OSIDisk = (unsigned char *) malloc (pHFE->number_of_track * nTrackSize);
	if (!OSIDisk)
	{
		LogText( _T("Error allocating target disk image memory"));
		goto abort;
	}

	memset(OSIDisk, 0, pHFE->number_of_track * nTrackSize);
	trkLst = (pictrack*) (pHFEDisk + pHFE->track_list_offset * 0x200);

	

	if ((unsigned char *) &trkLst[0] < pEnd)
	{

		for (int i = 0; i < pHFE->number_of_track; i++)
		{

			unsigned char * pData = pHFEDisk + (trkLst[i].offset * 0x0200);
			int pos = (pData - pHFEDisk);
			int nDataLen = trkLst[i].track_len;
			int nOutPos = 0;

			int loop = 0;
			if (bVerbose)
				LogText( _T("\r\nTrack %d @ $%04lX-$%04lX ($%04lX long)\r\n--------\r\n"),i,trkLst[i].offset * 0x0200, trkLst[i].offset * 0x0200 + trkLst[i].track_len, trkLst[i].track_len);

			//new track, reset decoder
			FMdecodeInitialize(fm);
			unsigned char * pOut = OSIDisk + i * nTrackSize;

			while (loop < nDataLen)
			{
				unsigned char * inBuf = pData + loop;
				int nInLen = 256;
				if (nInLen + loop > nDataLen)
				{
					nInLen = nDataLen - loop;
				}
				pOut = OSIDisk + i * nTrackSize + nOutPos;
				int nOutSize = nTrackSize;
				int nRead = 0;
				int nWrote = 0;

				FMdecodestream(inBuf, nInLen, pOut, nOutSize,fm, &nRead, &nWrote, pos+loop, bVerbose);

				if (nWrote > 0)
					nOutPos += nWrote;

				if (nWrote >= nOutSize)
				{
					//handle full output buffer
				}

				//input data is stored in 0x100 chunks, alternating between side 1 & 2
				//we only want 1 side's data
				loop += (pHFE->number_of_side ? pHFE->number_of_side * 256 : 256  );

			
			}
			if (fm.nZeroRun > 20)
			{
				if (bVerbose)
					LogText( _T("\r\n Zeros %02d \r\n"),fm.nZeroRun);
			}
			if (dt == DISK_UNKNOWN )
			{
				pOut = OSIDisk + i * nTrackSize;
				int offset =  OSISeekTrackStart( pOut, nOutPos);
				if (offset >= 0)
				{
					if (pOut[offset] == 0x43 && pOut[offset+1] == 0x57)
					{
						if (pOut[offset+2] < 0x80)
						{
							int nTrk = BCD2BIN(	pOut[offset+2] );
							if (abs(nTrk - i) <= 3)
							{
								nTrkOffset = i-nTrk;
								if (pHFE->number_of_track > 40)
									dt = OS65D8;
								else
									dt = OS65D5;
							}
							if (i == 0 && (dt == OS65D8 || dt == OS65D5)) //OSI sector on tk0, probably DOS/65
							{
								OSISector sect;
								memset(&sect, 0, sizeof(OSISector));
								
								if (dt == OS65D5 && OSICountSectors( pOut, nOutPos , sect) == 1)
								{
									if ( sect.nBlockLen[0] == 8 )  //DOS65 always single sector of 8 or 13 blocks
										dt = DOS655;
									else if (sect.nBlockLen[0] == 13)
										dt =  DOS658;
								}
							}
						}
					}
					else if ( pOut[offset] == 0x57 && pOut[offset+1] == i )
					{
						if (pHFE->number_of_track <= 40)
							dt = HEXDOS;
					}

					else if ( pOut[offset] == 0x80 && pOut[offset+1] == 0x00  )
					{
						//we've found OS68U disk type which require different decoding
						int nTrk = pOut[offset+2];
						if (abs(nTrk - i) <= 3)
						{
							nTrkOffset = i-nTrk;

							dt = OS65U8;
						}
					}
				}



			}
			if (i > 0 && dt == OS65U8)
			{
				//we have OS68U disk type which require different decoding, start over
				//restart decode of mixed 8E1/8N1
				FMdecodeInitialize(fm);

				bool bHeader = true;
				loop = 0;
				nOutPos = 0;
				pData = pHFEDisk + (trkLst[i].offset * 0x0200);
				pos = (pData - pHFEDisk);
				nDataLen = trkLst[i].track_len;
				memset(OSIDisk + i * nTrackSize, 0, nTrackSize);

				while (loop < nDataLen)
				{	

					unsigned char * inBuf = pData + loop;
					int nInLen = 256;	 

					if (nInLen + loop > nDataLen)
					{
						nInLen = nDataLen - loop;
					}

					pOut = OSIDisk + i * nTrackSize + nOutPos;
					int nOutSize = bHeader ? 3 : nTrackSize; //decode 1st 3 bytes
					int nRead = 0;
					int nWrote = 0;

					FMdecodestream(inBuf, nInLen, pOut, nOutSize-nOutPos,fm, &nRead, &nWrote, pos+loop, bVerbose);

					if (nWrote > 0)
						nOutPos += nWrote;
					
					if ( nOutPos >= nOutSize)
					{
						if (bHeader)
						{
							FMdecodeInitialize(fm);
							fm.eSformat = s8N1;

							bHeader = false;
							nOutSize = nTrackSize-nOutPos;
							pOut += nWrote;	
							nWrote = 0;
							inBuf += nRead;
							
							//decode rest of input buffer
							FMdecodestream(inBuf, nInLen-nRead, pOut, nOutSize,fm, &nRead, &nWrote, pos+loop+nRead, false);		
							if (nWrote > 0)
								nOutPos += nWrote;
						}
						else
						{
							//handle full output buffer
							break;

						}
					}

					//input data is stored in 0x100 chunks, alternating between side 1 & 2
					//we only want 1 side's data
					loop += (pHFE->number_of_side ? pHFE->number_of_side * 256 : 256  );

				
				} //while decoding this track

			} //if OS65U

		} //for there are tracks

	}
	if (nTrkOffset > 0)
	{
			if (bVerbose)
				LogText( _T("Track Data skewed by %d tracks\r\n"),nTrkOffset);
		
		memmove(OSIDisk, OSIDisk+(nTrkOffset*nTrackSize), (pHFE->number_of_track-nTrkOffset)*nTrackSize);
	}

	if (ppOSI)																		   
		*ppOSI = OSIDisk;
	else
		free (OSIDisk);

	if (pnOSILen)
	{
		int nSize = pHFE->number_of_track;
		if (nSize > 77)
			nSize = 77;
		*pnOSILen = nSize * nTrackSize;
	}

	return true;
abort:

	if (OSIDisk)
	{
		free(OSIDisk);
		OSIDisk = NULL;
	}
	if (ppOSI)
		*ppOSI = NULL;
	if (pnOSILen)
		*pnOSILen = -1;

	return false;
}

//Lets clean an OSI disk Image in memory

// pDisk is source OSI disk memory image
// dt specifies source format and desired target size/type
// 
// ppData is returned OSI disk in memory (caller must free)
// pnLen returns size of OSI image in memory
//
// return true if OK
bool CleanOSIImage( unsigned char * pDisk, DISKTYPE dt, unsigned char** ppData, int* pnLen )
{
	int nTracks = 0; //40 for 5.25, 77 for 8"
	int nTrackLen = 0;


	switch (dt)
	{
		case OS65D5:
		case HEXDOS:
		case DOS655:
			nTracks = MAXTRK5+1;
			nTrackLen = TRKSIZ5;
			break;			 
		case OS65D8:
		case OS65U8:
		case DOS658:
			nTracks = MAXTRK8+1;
			nTrackLen = TRKSIZ8;
			break;
		default:
			return false;
	}

	unsigned char * pTrackBuf = (unsigned char *) malloc(nTrackLen * nTracks); //buffer for encoding track

	if (!pTrackBuf)
		return false;
	memset(pTrackBuf, 0, nTrackLen * nTracks); //zero track data
	  
	if (pnLen)
	{
		*pnLen = nTrackLen * nTracks;
	}

	if (pDisk)
	{
		int nTrack;
		OSISector sect;
		memset(&sect, 0, sizeof(OSISector));
		bool bDos65Tk0 = false;
		
		if (dt == OS65D5 && OSICountSectors( pDisk, nTrackLen , sect) > 0)
		{
			//normal OS65D disks do not have a regular track/sector on tk0
			bDos65Tk0 = true;
			if (dt == OS65D5) dt = DOS655;
			if (dt == OS65D8) dt = DOS658;
		}
		LogText( _T("Clean: disk type %s\r\n"),  szIdentity[dt]);


		for ( nTrack = 0; nTrack < nTracks; nTrack++)
		{
			unsigned char * pDestTrack = pTrackBuf + nTrack * nTrackLen;
			unsigned char * pSrcTrack = pDisk + nTrack * nTrackLen;

			int nSectors = OSICountSectors( pSrcTrack, nTrackLen , sect);
			if (nTrack == 0 && nSectors == 0)
			{
				//just copy whole track, some OS use 8,9 or 13 blocks
				memcpy(pDestTrack, pSrcTrack, nTrackLen);				
			}
			else
			{
			 
				if (dt == HEXDOS)
				{
					//write track id
					*pDestTrack++ = 0x57;
					*pDestTrack++ = (unsigned char) nTrack;
					 
					if (nSectors > 0) 
					{
						const unsigned char * p = sect.pData[0];

						//write sector data
						for (int j = 0; j < sect.nBlockLen[0]; j++)
						{
							*pDestTrack++ =*p++;				
						}

					 }
					 else
					 {		//no sectors, unknown or blank formatting

						const unsigned char * p = pSrcTrack ? pSrcTrack+2 : NULL; //skip track marker

						//write unknown track data, no special formatting
						for (int j = 0; j <  2048; j++)
						{
							*pDestTrack++ =*p++;

						}

					 }
				}
				else if (dt == OS65D5 || dt == OS65D8 || dt == DOS655 || dt == DOS658)
				{
					int nCount = 0;
					
					//write track leader
					*pDestTrack++ = 0x43;
					*pDestTrack++ = 0x57;
					*pDestTrack++ = (unsigned char)BIN2BCD(nTrack);
					*pDestTrack++ = 0x58;
		
					if (nSectors > 0) 
					{
						for (nCount = 0; nCount < nSectors; nCount++)
						{
						
							//write sector header
							*pDestTrack++ = 0x76;
							
							if (sect.nBlockLen[nCount] < 0)
							{
								//write broken sector data from DOS65
								const unsigned char * p = sect.pData[nCount];

								//write sector data
								for (int j = 0; j < -sect.nBlockLen[nCount]; j++)
								{
									*pDestTrack++ = *p++;
								
								}

							}
							else
							{
							
								*pDestTrack++ = (unsigned char)(nCount+1);
								*pDestTrack++ = (unsigned char)sect.nBlockLen[nCount];
								const unsigned char * p = sect.pData[nCount];

								//write sector data
								for (int j = 0; j < sect.nBlockLen[nCount] * 256; j++)
								{
									*pDestTrack++ = *p++;
								
								}
								//write end of sector marker
								*pDestTrack++ = 0x47;
								*pDestTrack++ = 0x53;
							}


						}

					}
					else 
					{	
						//no sectors? just copy entire remaining track data
						int i = OSISeekTrackStart(pSrcTrack, nTrackLen);
						if (i >= 0)	//index of $43
						{
							i+=4; //skip track header
							memcpy(pDestTrack, pSrcTrack+i, nTrackLen-i);
							LogText( _T("Empty Track: %02d\r\n"),  nTrack);

						}
						else
						{
							LogText( _T("No Track Header on track: %02d\r\n"),  nTrack);

							memcpy(pTrackBuf + nTrack * nTrackLen, pSrcTrack, nTrackLen);
						}
						
					}
				}
				else if (dt == OS65U8)
				{	//OS65U
					//one 3584 byte sector, encoding switches mid-track
					//find start of sector data	within a reasonable about of bytes
					int j = OSISeekTrackStart(pSrcTrack, nTrackLen, dt);

					if (j >= 0 && pSrcTrack[j] == 0x80)
					{
						j +=3; //skip track start

						if ( pSrcTrack && (pSrcTrack[j] & 0x80) ) //skip possible noise byte, usually $FF, next data usually 0 or 1F
							j++;    
					
						//write track header
						*pDestTrack++ = 0x80;
						*pDestTrack++ = 0x00;
						*pDestTrack++ = (unsigned char)nTrack;
						memcpy(pDestTrack, pSrcTrack+j, 0xE03);	//sector data + chksum
						int chksum = 0;
						int i;
						for (i = 0; i <= 0xE03; i++)
						{
							//sum including track header
							unsigned char c = pDestTrack[i-3];
							chksum += c;						

						}
						if (pDestTrack[0xE02] != (chksum & 0xFF)  || pDestTrack[0xE01] != ((chksum >> 8) & 0xFF))
						{
							LogText( _T("Chksum fail on track: %02d\r\n"),  nTrack);

						}
						//


					}
					else
					{
						LogText( _T("No Track Header on track: %02d\r\n"),  nTrack);
						memcpy(pTrackBuf + nTrack * nTrackLen, pSrcTrack, nTrackLen);
					}

				}
				else
				{	//unknown disk format?
					memcpy(pTrackBuf + nTrack * nTrackLen, pSrcTrack, nTrackLen);
				}
			}

		}
		if (ppData)
		{
			*ppData = pTrackBuf;
		}
		else
		{
			free(pTrackBuf);
		}

		return true;

	}
	return false;


}

TCHAR * GetOSIDirectory(unsigned char * pMem, DISKTYPE dt)
{
	if (!pMem)
		return NULL;
	
	switch (dt)
	{
		case OS65D5:
		case OS65D8:
			return GetOS65DDir(pMem, dt);
			break;

		case HEXDOS:
			return GetHEXDOSDir(pMem, dt);
			break;

		case DOS655:
		case DOS658:
			return GetDOS65Dir(pMem, dt);
			break;			 
		
		case OS65U8:
			return GetOS65UDir(pMem, dt);
			break;
		default:
			return false;
	}

}

// OS65D directory
//
// 2 sectors (1 & 2) usually on track 12 (8 for 8") hold the directory information. 
// Each entry requires 8 bytes. There are a total of 64 entries. 
// The entries are formatted as follows:
//  Bytes 0-5 ASCII 6 character filename
//  Byte    6 BCD first track of file
//  Byte    7 BCD Last track of file

// TCHAR * GetOS65DDir(unsigned char * pTrack, DISKTYPE dt)
//
// pTrack is OS65D directory track (12 on 5.258 on 8")
// Return text of directory, memory must be freed by caller
//
#define BUFSIZE 8192
TCHAR * GetOS65DDir(unsigned char * pMem, DISKTYPE dt)
{
	
	if (!pMem || (dt != OS65D8 && dt != OS65D5))
		return NULL;			

	
	TCHAR * pBuf = NULL;
	unsigned char * pTrack = pMem;

	//dir offset:
	// $43, $57 - track start
	// $12 - track #
	// $58, 
	// $76,- sector start
	// $01, $01 - sector#, sector length
	// Entry 6 bytes = name, trk lo, trk hi (* 32 entries)
	//		(unused entry is "######', 0, 0)
	// $47, $53, 
	// $76,- sector start 
	// $02, $01 2nd sector start & length  (* 32 entries)
	//
	OSISector sect;
	int nLen = 0;
	int nMaxTrack = 0;
	
	switch (dt)
	{
		case OS65D5:
			pTrack = pMem + DIRTRACK5 * TRKSIZ5;
			nLen = TRKSIZ5;
			nMaxTrack = 40;
			
			break;

		case OS65D8:
			nLen = TRKSIZ8;
			nMaxTrack = 77;//77
			pTrack = pMem +	DIRTRACK8 * TRKSIZ8;
			break;
		default:
			break;
		
	}

	unsigned char szPico[] =  "PICO-DOS " ;

	if (bStrStr( pMem, szPico , 256, sizeof(szPico)-1 ) )
	{	
		//no dir on Pico Dos
		pBuf = (TCHAR *) malloc( 240);
		if (pBuf)
			_tcscpy( pBuf, _T("Pico Dos\r\n--------\r\n"));
		return pBuf;
	}
	unsigned char szCPM[] =  "OSI CP/M" ;

	//look for CPM marker in 1st 3 tracks
	if (dt == OS65D8 && bStrStr( pMem, szCPM , 3 * TRKSIZ8, sizeof(szCPM)-1 ) )
	{
		unsigned char *pTrack = pMem +	3 * TRKSIZ8;
		int nSectors = OSICountSectors(pTrack, nLen, sect);
		
		//This information is based on one sample disk and CPM 1.4 manuals
		//CPM assumes max 64 files per disk  (or not in the case of OSI CPM 1.4???)	
		//occupies trk 0 record 1-16 (2K) (or 3K for OSI?)

		//File Control Block 33 byte (disk FCB = 32 bytes)
		//0 =  entry type (0) not used (or 1-16 drive#)	for us 0=entry is valid
		//1-8  filename	padded with spaces
		//9-11 extension padded with spaces
		//12   file extent (normally 0)
		//13-14 not used (normally 0)
		//15   record count is current extent size 0 to 128 records
		//16-31 disk allocation disk block map (1 block = 1K or 8 records of 128 bytes
		//32 next record
		//there may be multiple directory entries with the same name if the file > 16K
		//the extent is incremented for each 16K entry

		//A single density disk has a block size of 1K, eack block holds 8 128byte sectors
		//CPM directory occupies the 1st one or more data blocks on the disk.  
		//Since directory on OSI CPM 1.4 is on track 3 the logical track # 0 = physical track # 3 (? probably).

		//Directory entry examples:
		//00 53 54 41 54 20 20 20 20 43 4F 4D 00 00 00 18 03 04 05 00 00 00 00 00 00 00 00 00 00 00 00 00
		//vv  S  T  A  T              C  O  M sq       rc -----------------------------------------------
		// vv->0=valid    rc = # of records occupied         a table that holds the # of each consecutive
		//                     by file                       1K blocks on the disk that holds this file
		//  sq sequence # used if file > 16K 
		// STAT.COM is 24 records long, found on disk blocks 03-05 		goes to trk 4
		// Since each track is 26 sectors (of 128 bytes) each
		// starts blk 3 = 3*8 = 0r24 + trk base = 3r24
		// ends blk 5 = 5*8 = 1r14 + trkbase =  4r14  +8? = 4r22

		//00 50 49 50 20 20 20 20 20 43 4F 4D 00 00 00 38 06 07 08 09 0A 0B 0C 00 00 00 00 00 00 00 00 00 
		//    P  I  P                 C  O  M		   rc
		//  PIP.COM = 56 sectors 7K	disk blocks 6-12	 Goes to trk 4,5,6
		// starts 6*8=48/26 = 1r22 + track base = trk4 rec 22 (offset 0xB00)
		// ends 12*8=96/26 =	3r18 + trk base = track 6, rec 18 (offset 0x0900) + 8? = rec26

		//00 41 53 4D 20 20 20 20 20 43 4F 4D 00 00 00 40 9D 9E 9F A0 A1 A2 A3 A4 00 00 00 00 00 00 00 00 
		//    A  S  M                 C  O  M		   rc
		// ASM.COM 64 sectors, 8K disk blocks 157-164	Goes to trk 51,52,53
		// start block 157 * 8 = 1256 / 26 = 48r8 48*26=1248 1256-1248=8 +trk base 3= TRK 51 record +8 (offest 0x400)	
		// end block 164 * 8 = 1312 / 26 = 50r12 + trk base = 53 rec 12 +8= 53 rec 20 (offset 0xa00) 

		//00 43 38 30 20 20 20 20 20 43 4F 4D 00 00 00 80 02 13 14 15 16 17 1A 1B 20 21 48 49 4A 4B 4C 4D 
		//00 43 38 30 20 20 20 20 20 43 4F 4D 01 00 00 6C 4E 4F 50 51 52 53 54 56 57 58 59 5A 5B 5C 00 00 
		//    C  8  0                 C  O  M sq       rc ------ block map ------------------------------
		//                                             ^- files size is sum of rc's 
		// block 2? trk 0r16 + trkbase = trk 3r16 to r24?
		// block 13 * 8 =104/ 26 = 4r0 + trk base = trk 7r0 

		//trk4-76 = 72 tracks *13*256 bytes = 239616 total size	1872 blks
		
		if (nSectors == 1 && sect.nBlockLen[0] == 13)
		{
			pBuf = (TCHAR *) malloc(BUFSIZE);
			
			if (!pBuf)
				return NULL;
			memset(pBuf, 0, BUFSIZE);
			TCHAR * pCurrent = pBuf;

			int nSize = 3328;  // OSI CPM 1.4 disk ?? recs go to 3200  should only go to 2048 corrupted disk?
			//make a copy of track to mark entries we process
			unsigned char *pCopy = (unsigned char *) malloc(nSize+1);	//64*32 entries	+ 40 more ???
			int nTotalBlks = 0;
			if (pCopy)
			{
				
				memcpy(pCopy, sect.pData[0], nSize); 
				_tcscpy(pCurrent, _T("OSI CPM\r\nRecs Bytes Ex Filename.typ\r\n--------------------------\r\n") );
				pCurrent += _tcslen(pCurrent);
				
				unsigned char *s = (unsigned char *) pCopy;
				unsigned char *pName, *pExt;
				char entry[32] = {0};
				while (s < (pCopy + nSize) )
				{
					int pos = 0;
					if (*s == 0 && *(s+1) > 32 && *(s+1)<= 'Z' )
					{
						
						memset(entry, 32, 31);
						int i;
						char c;
						pName = s+1;
						pExt = s+9;

						for (i = 0; i < 8; i++)
						{
							c = *(s+i+1);
							if (c > 0x20)
							{
								entry[pos++] = c;
							}

						}
						entry[pos++] = '.';
						int dot = pos;

						//It is said high bit set has the following meaning for the extension
						//characters T1-T3
						//T1-Read only, T2-system file, T3=archive
						for (i = 0; i < 3; i++)
						{
							c = *(s+i+9);
							if (c > 0x32)
								entry[pos++] = (char) (c&0x7F);

						}
						if (dot==pos)
							entry[--pos] = '\0';
						
						
						int blk = 0;
						int ext = 1;
						unsigned char *e = s+32;
						blk = (unsigned char) *(s+15);

						//now find all matching entries & total block sizes
						while ( e < (pCopy + nSize))
						{
							
						   if ( *e == 0 && !memcmp(pName, e+1, 8) && !memcmp(pExt, e+9, 3))
						   {
								   *e = 0xFF;
								   blk += *(e+15);
								   ext++;
						
						   }
							e  += 32;
						}
				
						entry[pos] = 0;

						if (*entry)
						{
#if (_MSC_VER >= 1400) && (defined(UNICODE))
							swprintf(pCurrent, _T("%4d %4dK %2d %S\r\n"), blk, (blk+7)/8,  ext,  entry );
#else							
							_stprintf(pCurrent,  _T("%4d %4dK %2d %S\r\n"), blk, (blk+7)/8,  ext,  entry );
#endif
							
							pCurrent += _tcslen(pCurrent);
							nTotalBlks += blk;
						}
					}
					s = s +32;


				}

				free(pCopy);
				pCopy = NULL;

			} //if pCopy

#if (_MSC_VER >= 1400) && (defined(UNICODE))
			swprintf(pCurrent, _T("%d blks of 1872, %dK Free\r\n"), nTotalBlks, (1872-nTotalBlks+4)/8 );
#else							
			_stprintf(pCurrent, _T("%d blks of 1872, %dK Free\r\n"), nTotalBlks, (1872-nTotalBlks+4)/8 );
#endif
			pCurrent += _tcslen(pCurrent);
		
			if (!nTotalBlks)
			{
				free(pBuf);
				pBuf = NULL;
			}
			else
			{
					return pBuf;

			}

		}
		
	}


	
	int nSectors = OSICountSectors(pTrack, nLen, sect);
	if (nSectors)
	{
		//14 character filename extension for OS65D
		//uses two file entries, only the 2nd contains track information
		//how can we determine if 14 character filenames are in use?
		// if track ranges are wrong, (ASCII) in BCD track locations

		pBuf = (TCHAR *) malloc(BUFSIZE);
		if (!pBuf)
			return NULL;

		TCHAR * pCurrent = pBuf;

		static TCHAR item[50];
		int nCount = 1;
		
		bool bTrackMap[80] = { 0 };
		bool bTrackMape[80] = {0};
		int trk_collision =0;
		int trke_collision =0; 

		//1st pass decide if valid normal or extended filenames
		int nExtended = 0;
		int nNormal = 0;
		for (int j = 0; j < __min(2, nSectors); j++)
		{
			const unsigned char * pSector = sect.pData[j];
			
			for (int entry = 0; entry < 32; entry++)
			{
				bool bEmpty = true;
				int i;
				bool bASCII = true;

				for ( i = 0; i < 6; i++)
				{
					
					unsigned char c = *(pSector + (8 * entry) + i);
					if (c != 0x23)
						bEmpty = false;
					if (!bEmpty && ( c < 0x20 || c > 0x7F))
						bASCII = false; 
				}


				if (!bEmpty && bASCII)
				{

					unsigned char s1 = *(pSector + 8 * entry + 6);
					unsigned char e1 = *(pSector + 8 * entry + 7);
					int start = BCD2BIN( s1 );
					int end   = BCD2BIN( e1 );
					//s1 & e1 could be ASCII in 14char filename or BCD track range in 6 char filename

					if (!IsValidBCD(s1) || !IsValidBCD(e1) || (end < start) || 
						(end == 0) || (start >= nMaxTrack) || (end >= nMaxTrack) ||
						(s1 >= 0x20 && s1 <= 0x7f && e1 >= 0x20 && e1 <= 0x7f) )
					{
						if ((entry & 1) == 0)
						{
							//if trk range looks like ascii, it may be extended
							if (s1 >= 0x20 && s1 < 0x80 && e1 >= 0x20 && e1 < 0x80)
							{
								unsigned char s2 = *(pSector + 8 * entry + 6+8);
								unsigned char e2 = *(pSector + 8 * entry + 7+8);
								int start2 = BCD2BIN( s2 );
							    int end2   = BCD2BIN( e2 );

								if (IsValidBCD(s2) && IsValidBCD(e2) &&	(end2 >= start2) && (end2 < nMaxTrack)
									&& (end2 != 0) &&  (start2 < nMaxTrack) )
								{
									nExtended++;
									for (int k = start2; k <= end2; k++)
									{
										if (bTrackMape[k])
										{
											trke_collision++;
										}
										else
											bTrackMape[k] = true;
									}

								}

							}
						}

					}
					if (IsValidBCD(s1) && IsValidBCD(e1) && (end >= start) && 
						(end != 0) && (start < nMaxTrack) && (end < nMaxTrack))
					{
						for (int k = start; k <= end; k++)
						{
							if (bTrackMap[k])
							{
								trk_collision++;
							}
							else
								bTrackMap[k] = true;
						}
						
						nNormal++;
					}

				
				}


			}
			


	   }
	   if (nExtended > 1 && nNormal > 1)
	   {
		if (trk_collision > trke_collision)
			nNormal = 0;
	   }


		if (nNormal > 0 && nNormal >= nExtended)
		{
			
			if (trk_collision > 5)
			{
				//invalid disk?	too many track use collisions
				free(pBuf);
				return NULL;

			}			
			
			_tcscpy(pCurrent, _T("OS65D File  Track range\r\n-----------------------\r\n") ); 
			
			for (int j = 0; j < __min(2, nSectors); j++)
			{
				const unsigned char * pSector = sect.pData[j];

				
				for (int entry = 0; entry < 32; entry++)
				{
					bool bEmpty = true;
					int i;

					for (i = 0; i < 40; i++)
						item[i] = _T(' ');
						 
					for ( i = 0; i < 6; i++)
					{
						
						TCHAR c = *(pSector + 8 * entry + i);
						item[i+1] = c;
						
						if (c != _T('#'))
							bEmpty = false;
					}


					if (!bEmpty)
					{
						unsigned char s1 = *(pSector + 8 * entry + 6);
						unsigned char e1 = *(pSector + 8 * entry + 7);

						if (IsValidBCD(s1) && IsValidBCD(e1))
						{
						
							int start = BCD2BIN( *(pSector + 8 * entry + 6) );
							int end   = BCD2BIN( *(pSector + 8 * entry + 7) );

							if ( end != 0 && start <= end )
							{
								nCount++;
#if (_MSC_VER >= 1400) && (defined(UNICODE))
								swprintf(&item[13], 10, _T("%02d - %02d\r\n"), start, end);
#else							
								_stprintf(&item[13],  _T("%02d - %02d\r\n"), start, end);
#endif
								_tcscat( pCurrent, item);
							}
						}
					
					}


				}
			}
			if (nCount)
			{
#if (_MSC_VER >= 1400) && (defined(UNICODE))
				swprintf(item, 40, _T("\r\n%02d entries free out of %02d\r\n"), 65-nCount, 64);	
#else
				_stprintf(item, /*40,*/ _T("\r\n%02d entries free out of %02d\r\n"), 65-nCount, 64);	
#endif
				_tcscat( pCurrent, item);

			}

		}

		if (nExtended > nNormal )
		{

			if (trke_collision > 5)
			{
				//invalid disk?
				free(pBuf);
				return NULL;
			}

			_tcscpy(pCurrent, _T("OS65De File       Track range\r\n-----------------------------\r\n") ); 
			
			for (int j = 0; j < __min(2, nSectors); j++)
			{
				const unsigned char * pSector = sect.pData[j];

				
				for (int entry = 0; entry < 16; entry++)
				{
					bool bEmpty = true;
					int i;

					for (i = 0; i < 48; i++)
						item[i] = _T(' ');
						 
					for ( i = 0; i < 14; i++)
					{
						
						TCHAR c = *(pSector + 16 * entry + i);
						item[i+1] = c;
						
						if (c != _T('#'))
							bEmpty = false;
					}


					if (!bEmpty)
					{
						unsigned char s1 = *(pSector + 16 * entry + 14);
						unsigned char e1 = *(pSector + 16 * entry + 15);

						if (IsValidBCD(s1) && IsValidBCD(e1))
						{
						
							int start = BCD2BIN( *(pSector + 16 * entry + 14) );
							int end   = BCD2BIN( *(pSector + 16 * entry + 15) );

							if ( end != 0 && start <= end )
							{
								nCount++;
#if (_MSC_VER >= 1400) && (defined(UNICODE))
								swprintf(&item[21], 10, _T("%02d - %02d\r\n"), start, end);
#else							
								_stprintf(&item[21],  _T("%02d - %02d\r\n"), start, end);
#endif
								_tcscat( pCurrent, item);
							}
						}
					
					}


				}
			}
			if (nCount)
			{
#if (_MSC_VER >= 1400) && (defined(UNICODE))
				swprintf(item, 50, _T("\r\n%02d entries free out of %02d\r\n"), 33-nCount, 32);	
#else
				_stprintf(item, /*40,*/ _T("\r\n%02d entries free out of %02d\r\n"), 33-nCount, 32);	
#endif
				_tcscat( pCurrent, item);

			}


		}
			if (nCount <= 1)
			{
				free(pBuf);
				pBuf = NULL;
			}

		}

	return pBuf;

}

// Decode OS65U directory
//
// OS65U file contents: 1st 16 bytes contains copy of the file directory entry 
// (filename, attibutes, address etc.)
//
// OS65U Directory is stored in file Direc*, there are 16 bytes per file entry
// 6 bytes for filename, 
// 2 bytes password LO, HI,
// 1 byte  file attributes  
// 3 bytes starting pos ll, mm, hh (24 bit address)
// 3 bytes files size   ll, mm, hh (24 bit address)
// 1 bytes semaphore(?) 0x00
//
// Attributes byte has the following meaning
//           bit 7=dir flag
//           bit 6 ?
//           bit 5 ?
//           bit 4 - \
//           bit 3 - -- 3 bits file type: 0 = basic, 1 = data else 'other'
//           bit 2 - /
//           bit 1 \ __ 2 bits access:  0=none 1=Read, 2=Write, 3=R/W
//           bit 0 /
//
// End of directory has null filename 
// A deleted file has 0x01 for filename
// A password is entered as 4 chars uppercase ASCII, 'A' to ']'
// if 1st char is A-P then subtract 'A'(0x41) else = 0x0F  = c1
// if 2nd char is N-] then subtract 'N'(0x4E) else = 0x0F  = c2
// if 3rd char is A-P then subtract 'A' else = 0x0F	 = c3
// if 4th char is N-] then subtract 'N' else = 0x0F	 = c4
// PL = c1 << 4 | c2
// PH = c3 << 4 | c4
// so password 'PASS' = F, F, F, 5 : PL = FF, PH = F5 on directory
// password 'ANAN' = 0, 0, 0, 0
// DIREC* file specifies size of directory in it's length (size) field in pages. 
// A whole track is 0E length,  partial track could be '04' length, leaving space for other files on that disk.
//
// OS65U occupies 1st 25088 (0x6200) bytes or 7 sectors
// The first file is always DIREC*, and since a sector is 3584 bytes and there are 16 bytes per directory entry, 
// a floppy disk can hold the DIRECT* entry and AT MOST 223 additional entries. Smaller sector sizes allow for 
// fewer directory entries.
// --------
//
// TCHAR * GetOS65UDir(unsigned char * pTrack, DISKTYPE dt)
//
// Input: OS65U directory track, and Disk Type (OS65U8)
// Returns contents of directory as text
// Caller is responsible for deleting returned memory
//
TCHAR * GetOS65UDir(unsigned char * pMem, DISKTYPE dt)
{

	if (!pMem || dt != OS65U8)
		return NULL;
	
	unsigned char* pTrack = pMem + TRKSIZ8 * DIRTRACK65U;
	
	TCHAR item[128] = {0};
	TCHAR pass[5] = {0};

	int nCount = 0, nPos; 
	int nPad = 0;
	unsigned char key[4] = { 0x80, 0x00, 0x07 };
	unsigned char *pCur = pTrack, *pStart = pTrack;
	unsigned char pwLo, pwHi;
	unsigned char cAttrib;
	int nStart, nSize;
	unsigned char cSemi;
	const TCHAR * pAccess = _T("");
	const TCHAR * pType = _T("");
	TCHAR *pList = NULL;
	int nNotOS65UDir = 0;
	int nAddrMax = 0;
	int nEntry = 0;
	int nRecover = 0;
	int nTotalEntries = 0;

	if (!pStart)
		return NULL;
	
	if (dt != OS65U8)
		return NULL;

	//determine start of directory data	on track 7
	for (nPad = 0; nPad < 5; nPad++)
	{
		if (!memcmp(pStart+nPad, key, 3) )
			break;
	}
	if (nPad == 5 )
		return NULL;

	pStart = pTrack + nPad+3;
	
	//skip possible $FF and $1F
	for (nPad = 0; nPad < 5; nPad++)
	{
		if (*(pStart+nPad) == 0x1F)
			break;
		if (*(pStart+nPad) == 0x0F)
			break;
		if (*(pStart+nPad) == 0x44)	//1st character is the contents of the files directory entry
		{
			nPad--;
			break;
		}


	}
	if (nPad == 5 ) 
		return NULL;
	
	pList = (TCHAR*) malloc (16384);
	if (!pList)
		return NULL;
		
	pStart += nPad+1;
	if (!strncmp((char *) pStart, "DIREC*", 6))	 //skip file header entry
	{
		pCur = pStart+9;
		nStart = (*pCur << 8) + (*(pCur+1) << 16) + (*(pCur+2) << 24);
		pCur += 3;
		nSize = (*pCur << 8) + (*(pCur+1) << 16) + (*(pCur+2) << 24);
		nAddrMax = nStart + nSize;
		nTotalEntries = nSize/16-1;
		pStart += 16;
	}

	pCur = pStart;

	_tcscpy(pList,
	// "abcdef Othr2  R/W
	_T("Name    Type  Access Address/Track Length / Sec SecBnd SecLen PassWord\r\n")
	_T("----------------------------------------------------------------------\r\n") );
						

	while (*pCur && pCur < pTrack + 0xE00 + nPad)
	{
		nEntry++;
		pCur = pStart + 16*nCount;
		
		//copy 6 byte file name
		for (nPos = 0; nPos < 6; nPos++)
		{
			item[nPos] = *(pCur+nPos);
			if ( (item[nPos] > 0x01 && item[nPos] < 0x20) || (item[nPos] & 0x80) )	 //not really an OS65U name?
				nNotOS65UDir++;
		}
		item[6] = 0;
		pCur += 6;
		pwLo = *pCur++;
		pwHi = *pCur++;

		cAttrib = *pCur++;

		nStart = (*pCur << 8) + (*(pCur+1) << 16) + (*(pCur+2) << 24);
		pCur +=3;
		nSize = (*pCur << 8) + (*(pCur+1) << 16) + (*(pCur+2) << 24);
		pCur +=3;

		if (nAddrMax < nStart + nSize)
			nAddrMax = nStart + nSize;

		cSemi = *pCur++;
		//determine file access
		switch (cAttrib & 0x03)
		{
			case 0:
				pAccess = _T("None ");
				break;
			case 1:
				pAccess = _T("Read ");
				break;
			case 2:
				pAccess = _T("Write");
				break;
			case 3:
				pAccess = _T(" R/W ");
				break;
		}
		//determine file type
		switch ((cAttrib >> 2) & 0x07)
		{
			case 0:
				pType = _T("Data ");
				break;
			case 1:
				pType = _T("Basic");
				break;
			case 2:
				pType = _T("Othr2");
				break;
			case 3:
				pType = _T("Othr3");
				break;
			case 4:
				pType = _T("Othr4");
				break;
			case 5:
				pType = _T("Othr5");
				break;
			case 6:
				pType = _T("Othr6");
				break;
			case 7:
				pType = _T("Othr7");
				break;
		}

		//high bit is directory flag
		if (cAttrib & 0x80)
			pType = _T("Direc");

		//decode file password.... shhh this is (supposedly) a secret
		if (pwLo == 0xFF && pwHi == 0xF5)
			_tcscpy(pass, _T("PASS"));
		else
		{
			pass[0] = (TCHAR) (_T('A') + (pwLo>>4));
			pass[1] = (TCHAR) (_T('N') + (pwLo & 0x0F));
			pass[2] = (TCHAR) (_T('A') + (pwHi>>4));
			pass[3] = (TCHAR) (_T('N') + (pwHi & 0x0F));
			pass[4] = 0;
		}
		bool bSec = nStart/3584*3584 == nStart;
		bool bLen = nSize/3584*3584 == nSize;
	    
		//item 
		if (*item == 0x01)
		{
			_tcscpy((TCHAR*) item, _T("(del) "));
			nRecover += nSize;
		}
		nCount++;
#if (_MSC_VER >= 1400) && (defined(UNICODE))
		swprintf( ((TCHAR *) &item)+6, 120, _T("  %s  %s %6d / %02d   %6d / %02d  %s    %s    %s\r\n"), pType,pAccess, nStart, nStart/3584, nSize, (nSize+3583)/3584,
		bSec ? _T("yes") : _T("no "), bLen ? _T("yes") : _T("no "), &pass[0] );
#else
		_stprintf( ((TCHAR *) &item)+6, /*120,*/ _T("  %s  %s %6d / %02d   %6d / %02d   %s    %s    %s\r\n"), pType,pAccess, nStart, nStart/3584, nSize, (nSize+3583)/3584,
		bSec ? _T("yes") : _T("no "), bLen ? _T("yes") : _T("no "), &pass[0] );
#endif
		_tcscat( pList, (TCHAR *)item);

	}
	
	_tcscat( pList, _T("\r\n"));
	if (nRecover)
	{
		_stprintf( item, _T(" %d Bytes Recoverable\r\n"), nRecover );
		_tcscat( pList, (TCHAR *)item);
	}

	if (nAddrMax <= 275968) //max 8" 77 trk floppy storage
	{
		int nFree = (275968 - nAddrMax);
		if (nFree >= 0)
		{
			_stprintf( item, _T(" %d Bytes Free\r\n %d Sector(s) Free\r\n"),nFree, nFree/3584);
			_tcscat( pList, (TCHAR *)item);

		}
	}
	if (nTotalEntries)
	{
			_stprintf( item, _T(" %d Files Defined of %d Possible\r\n"), nEntry, nTotalEntries);
			_tcscat( pList, (TCHAR *)item);
	
	}
	if (nNotOS65UDir > 10)
	{
		//too many invalid characters ... this can't be a real directory
		free(pList);
		pList = NULL;
	}



	return pList;
}
/* HEXDOS dir loads to $0B01
Stored as linked BASIC listing, line number is track number
Filenames can be variable length
A machine language file starts with $, the 1st two bytes are load & start address, lo byte 1st

        $0B01
         v
57 01    0C 0B 02 00 48 45 58 41 53 4D 00 12 0B 09 00 A5 00 00 00
TRK 01   ----- ~~~~~ =================    ----- NxtTrk *    -----
            |  Trk#   H  E  X  A  S  M     ^  |              ^
            |                              |  |              |
            --------------------------------  ----------------
Since the directory is stored as a BASIC file, Keywords are tokenized in directory listings
so any characters with high bit set are tokenized keywords.
* denotes 1st free track

*/
//
// TCHAR * GetHEXDOSDir(unsigned char * pTrack, DISKTYPE dt)
//
// Input: Hexdos directory track, and Disk Type 
// Returns contents of directory as text
// Caller is responsible for deleting returned memory
//
TCHAR * GetHEXDOSDir(unsigned char * pMem, DISKTYPE dt)
{

	if (!pMem || dt != HEXDOS)
		return NULL;

	unsigned char* pTrack = pMem + TRKSIZ5;


	// OSI BASIC in ROM	token list
	const TCHAR *keywordROM[] = 
	{
		_T("END"),  _T("FOR"),  _T("NEXT"), _T("DATA"),   _T("INPUT"),  _T("DIM"),    _T("READ"),  _T("LET"),
		_T("GOTO"), _T("RUN"),  _T("IF"),   _T("RESTORE"),_T("GOSUB"),  _T("RETURN"), _T("REM"),   _T("STOP"),
		_T("ON"),   _T("NULL"), _T("WAIT"), _T("LOAD"),   _T("SAVE"),   _T("DEF"),    _T("POKE"),  _T("PRINT"),
		_T("CONT"), _T("LIST"), _T("CLEAR"),_T("NEW"),    _T("TAB("),   _T("TO"),     _T("FN"),    _T("SPC("),   
		_T("THEN"), _T("NOT"),  _T("STEP"), _T("+"),      _T("-"),      _T("*"),      _T("/"),     _T("^"),  
		_T("AND"),  _T("OR"),   _T(">"),    _T("="),      _T("<"),      _T("SGN"),    _T("INT"),   _T("ABS"),
		_T("USR"),  _T("FRE"),  _T("POS"),	_T("SQR"),    _T("RND"),    _T("LOG"),    _T("EXP"),   _T("COS"),
		_T("SIN"),  _T("TAN"),  _T("ATN"),  _T("PEEK"),   _T("LEN"),    _T("STR$"),   _T("VAL"),   _T("ASC"),
		_T("CHR$"), _T("LEFT$"),_T("RIGHT$"),_T("MID$")

	};

	TCHAR * pList = NULL;
	if (pTrack)
	{
		//dir offset:
		// $57, $01 - track start
		// <2K sector data>
		int nPos = OSISeekTrackStart(pTrack, TRKSIZ5);

	    if (nPos >= 0)
		{
			int base = 0x0B01;
			if (pTrack[nPos] != 0x57 && pTrack[nPos+1] != 0x01)
				return NULL;
			nPos+=2;
			unsigned char* p = pTrack+nPos;

			int offset = 0;
			TCHAR szLine[1024];

			pList = (TCHAR*) malloc (16384);
			if (!pList)
				return NULL;

			_tcscpy(pList,
			_T("HEXDOS Track, Filename\r\n")
			_T("----------------------\r\n") );
 

			while (pList)
			{
				*szLine = 0;
				offset = *p + (*(p+1) << 8);
				if (offset == 0)
					break;
				int nTrack = *(p+2) + (*(p+3) << 8);
				if (nTrack < 40)
				{
#if (_MSC_VER >= 1400) && (defined(UNICODE))
					swprintf( szLine, _T("%02d, "), nTrack );
#else
					_stprintf( szLine, _T("%02d, "), nTrack );
#endif				
					
					unsigned char * pText = p+4;
					while (*pText)
					{
						if ( (*pText) & 0x80)
						{
							//keyword
							int i = *pText & 0x7F;
							if (i <	sizeof(keywordROM)/sizeof(keywordROM[0]))
							{
								_tcscat(szLine, keywordROM[i]);
							}
						}
						else
						{
							TCHAR buf[2] = {0};
							buf[0] = *pText; 
							_tcscat(szLine, buf);	
						}
						pText++;
					}
					_tcscat(szLine, _T("\r\n"));

				}
				_tcscat(pList, szLine);

				if (offset - base > 2048 || offset < base)
				{
					break;	
				} 
				//point to next basic line
				unsigned char* pNext = pTrack+nPos+ (offset - base);
				if ( pNext > p)
					p = pNext;
				else
					break; //reverse traverse?
			}

		}

		
	}
	return pList;

}

// Dos/65-5 store directory on track 4, the start or the data area
// each Dos65 Record is 128 bytes, 4 dir entries per record, generally 64 entries per disk
//
// Each entry is 32 bytes
// 52 55 4E 20 20 20 20 20 43 4F 4D 00 00 00 54 2D 2E 2F 30 31 32 33 34 45 36 37 00 00 00 00 00 00
// ~~~~~~~~~~~~~~~~~~~~~~~ ========	EE xx xx RR	BB BB BB BB BB BB BB BB BB BB BB BB BB BB BB BB NN  
//      (filename 0-7)    (extn 8-A)
// EE= Extent (0-$1F), RR = # Records BB= Block Numbers, NN=Next record in binary
// Each extent references up to 16K bytes as individual 128 byte records
// 5.25: MaxBlk=71, 40 tracks, NumSYStemTRks=4, NumRECRD/trk=16, Logical sectors are 1 more than physical
// 8"  : MaxBlk=239, 77 tracks, NumSYStemTRks=3, NumRECRD/trk=26	 BLKSCD=0 (1K), MAXDIR=63
//
// Translation table:
// Logical, Physical
//  0          1
//  1          7
//  2         13
//  3         19
//  4         25
//  5          5
//  6         11
//  7         17
//  8         23
//  9          3
// 10          9
// 11         15
// 12         31
// 13          2
// 14          8
// 15         14
// 16         20
// 17         26
// 18          6
// 19         12
// 20         18
// 21         24
// 22          4
// 23         10
// 24         16
// 25         22   
//	    	  
TCHAR * GetDOS65Dir(unsigned char * pDisk, DISKTYPE dt)
{
	TCHAR * pList = NULL;
	if (pDisk)
	{
		OSISector sect;
		memset(&sect, 0, sizeof(OSISector));
		int nTrackLen = TRKSIZ5;
		if (dt == DOS658)
			nTrackLen = TRKSIZ8;
		
		memset(&sect, 0, sizeof(OSISector));

		unsigned char* pDirTrack = pDisk + ( nTrackLen * DOS65DIRTRACK5);
		if (dt == DOS658)
			pDirTrack = pDisk + ( nTrackLen * DOS65DIRTRACK8);

		
		int i = OSICountSectors(pDirTrack, nTrackLen, sect);
		if (i == 1)
		{
			//$43, $57, $nTrk, $58, xxxx, $76, $01, $08, $00
			const unsigned char* p = sect.pData[0];
			if (p)
			{
				p++; //skip leading null
				pList = (TCHAR*) malloc (16384);
				if (!pList)
					return NULL;
				_tcscpy(pList,
					_T("DOS/65  File     #rec\r\n")
					_T("----------------------\r\n") );

				int entry = 0;
				while (entry < 63)
				{
					TCHAR c;
					//filename.extn  size, trk?
					TCHAR filename[17];
					for (i = 0; i < 16; i++)
						filename[i] = 0x20;
					filename[16] = 0;

					int recs = *(p+0x0E);
					//0 #recs = invalid (deleted?) file; NUL bytes not nul? = invalid
					if (recs > 0 && *(p+0x0C) == 0 && *(p+0x0D) == 0)
					{
						//copy files name
						for (i=0; i < 8; i++)
						{
							c = *(p+i);
							if ( c >= 0x20)
								filename[i] = c;
							else
								filename[i] = 0x20;
						}
						//copy extension
						int nExtn = 8;
						filename[8] = 0x2E;
						for (i = 0; i < 4; i++)
						{
							c = *(p+i+nExtn);
							if ( c > 0x20)
								filename[9+i] = c;
							else
								filename[9+i] = 0x20;
						}
						
						TCHAR *buf = pList + _tcslen(pList);

#if (_MSC_VER >= 1400) && (defined(UNICODE))
							swprintf( buf, _T("%s %04d\r\n"), filename, recs );
#else
							_stprintf( buf, _T("%s %04d\r\n"), filename, recs );
#endif
					}
					if (  *(p+0x1F) != 0)
						break;

					p = p+ 32;
					entry++;


				
				}


			}
		} 


	}

	return pList;

}